"""This file generates heatmaps for top genes of particular class w.r.t same top genes in other classes."""

import os
from typing import Tuple, Union

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from scalr.analysis import AnalysisBase
from scalr.utils import EventLogger
from scalr.utils import read_data


class Heatmap(AnalysisBase):
    '''Class to generate a heatmap of top genes classwise.'''

    def __init__(self,
                 top_n_genes: int = 100,
                 save_plot: bool = True,
                 score_matrix_path: str = None,
                 top_features_path: str = None,
                 *args,
                 **kwargs):
        """Initialize class with shap arguments.

        Args:
            top_n_genes: top N genes for each class/label.
            save_plot: Where to save plot or show plot.
            score_matrix_path: path to score matrix.
            top_features_path: path to top features.
        """

        self.top_n_genes = top_n_genes
        self.save_plot = save_plot
        self.score_matrix_path = score_matrix_path
        self.top_features_path = top_features_path

        self.event_logger = EventLogger('Heatmap')

    def generate_analysis(self,
                          dirpath: str,
                          score_matrix: pd.DataFrame = None,
                          top_features: Union[dict, list] = None,
                          **kwargs) -> None:
        """A function to generate heatmap for top features.

        Args:
            score_matrix: Matrix(class * genes) that contains a score of each gene per class.
            top_features: Class-wise top genes or list of top features.
            dirpath: Path to store the heatmap image.
        """

        self.event_logger.heading2("Generating Heatmaps.")

        if isinstance(top_features, list):
            self.event_logger.info(
                "Generating heatmap for the same top genes across all classes as provided"
                " `top_features` is a single list and not top genes per class dict."
            )
            top_features = {"all_class_common": top_features}

        if (score_matrix is None) and (top_features is None):

            if not self.score_matrix_path:
                raise ValueError("score_matrix_path required.")

            if not self.top_features_path:
                raise ValueError("top_features_path required.")

            score_matrix = read_data(self.score_matrix_path)
            top_features = read_data(self.top_features_path)

        for class_name, genes in top_features.items():
            self.plot_heatmap(score_matrix[genes[:self.top_n_genes]].T,
                              f"{dirpath}/heatmaps", class_name)

        self.event_logger.info(f"Heatmaps stored at: {dirpath}/heatmaps")

    def plot_heatmap(self, class_genes_weights: pd.DataFrame, dirpath: str,
                     filename: str) -> None:
        """A function to plot a heatmap for top n genes across all classes.

        Args:
            class_genes_weights: Matrix(genes * classes) which contains
                                 shap_value/weights of each gene to class.
            dirpath: Path to store the heatmap image.
            filename: Heatmap image name.
        """

        os.makedirs(dirpath, exist_ok=True)

        sns.set(rc={'figure.figsize': (9, 12)})
        sns.heatmap(class_genes_weights, vmin=-1e-2, vmax=1e-2)

        plt.tight_layout()
        plt.title(filename)

        if self.save_plot:
            plt.savefig(os.path.join(dirpath, f"{filename}.svg"))
        else:
            plt.show()
        plt.clf()



from . import analysis
from . import data
from . import feature
from . import nn
from . import utils
from .data_ingestion_pipeline import DataIngestionPipeline
from .eval_and_analysis_pipeline import EvalAndAnalysisPipeline
from .feature_extraction_pipeline import FeatureExtractionPipeline
from .model_training_pipeline import ModelTrainingPipeline



"""This file contains an implementation of the logger in the pipeline."""

import logging


class FlowLogger(logging.Logger):
    """Class for flow logger.
    
    It logs high-level overview of pipeline execution in the terminal.
    """
    level = logging.NOTSET

    def __init__(self, name, level=None):
        if level:
            FlowLogger.level = level

        if not FlowLogger.level:
            FlowLogger.level = logging.INFO

        super().__init__(name, FlowLogger.level)

        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s : %(message)s')

        handler = logging.StreamHandler()
        handler.setLevel(FlowLogger.level)
        handler.setFormatter(formatter)
        self.addHandler(handler)


class EventLogger(logging.Logger):
    """Class for event logger. It logs detailed file-level logs during pipeline execution.
    """
    level = logging.NOTSET
    filepath = None

    def __init__(self, name, level=None, filepath=None, stdout=False):
        """Initialize required parameters for event logger."""
        if level:
            EventLogger.level = level

        if not EventLogger.level:
            EventLogger.level = logging.INFO

        super().__init__(name, EventLogger.level)

        if filepath:
            EventLogger.filepath = filepath

        if not EventLogger.filepath:
            handler = logging.NullHandler()
        else:
            handler = logging.FileHandler(EventLogger.filepath)

        formatter = logging.Formatter('%(message)s')

        handler.setLevel(EventLogger.level)
        handler.setFormatter(formatter)
        self.addHandler(handler)

        # If user wants to print logs to stdout
        if stdout:
            handler = logging.StreamHandler()
            handler.setLevel(EventLogger.level)
            handler.setFormatter(formatter)
            self.addHandler(handler)

    def heading(self, msg, prefix, suffix, count):
        """A function to configure setting for heading."""
        self.info(f"\n{prefix*count} {msg} {suffix*count}\n")

    def heading1(self, msg):
        """A function to configure setting for heading 1."""
        self.heading(msg, "<", ">", 10)

    def heading2(self, msg):
        """A function to configure setting for heading 2."""
        self.heading(msg, "-", "-", 5)



"""This file generates differential gene expression analysis using Linear mixed-effects model approach and stores the results."""
import multiprocessing
import os
from os import path
import pickle
import resource
import string
import traceback
from typing import Optional, Tuple, Union
import warnings

from anndata import AnnData
from anndata import ImplicitModificationWarning
import anndata as ad
from anndata.experimental import AnnCollection
from joblib import delayed
from joblib import Parallel
import matplotlib.pyplot as plt
import numpy as np
from numpy.linalg import LinAlgError
from pandas import DataFrame
import pandas as pd
import scanpy as sc
from scipy.optimize import OptimizeWarning
import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.stats.multitest import multipletests
from statsmodels.tools.sm_exceptions import ConvergenceWarning
from statsmodels.tools.sm_exceptions import HessianInversionWarning

from scalr import utils
from scalr.analysis import AnalysisBase
from scalr.utils import EventLogger
from scalr.utils import read_data


class DgeLMEM(AnalysisBase):
    '''Class to perform differential gene expression analysis 
    using Linear mixed effects model.'''

    def __init__(self,
                 fixed_effect_column: str,
                 fixed_effect_factors: list[str],
                 group: str,
                 celltype_column: str = None,
                 cell_subsets: list[str] = None,
                 min_cell_threshold: int = 10,
                 n_cpu: int = 6,
                 gene_batch_size: int = 1000,
                 coef_threshold: [float, int] = 0,
                 p_val: Union[float, int] = 0.05,
                 y_lim_tuple: Optional[Tuple[float, ...]] = None,
                 save_plot: bool = True,
                 stdout: bool = False):
        '''DgeLMEM parameters initialization.
        
        Args: 
            fixed_effect_column: Column name in `anndata.obs` containing different factor levels or categories for
                           differential gene expression analysis. This acts as a fixed_effect parameter.           
            fixed_effect_factors: List of conditions in `fixed_effect_column` to make design matrix for.
            group: Column name to act as a random_effect parameter for mixed effect model.
            celltype_column: Column name in `anndata.obs` containing all the cell types.
            cell_subsets: Selcted list of cell types in 'celltype_column' to subset the anndata.
            min_cell_threshold: Minimum number of cells with nonzero values for a gene. Used for filtering noisy genes.
            n_cpu: Number of CPUs for parallelization.
            gene_batch_size: Number of genes in a batch of process.
            coef_threshold: Threshold to filter up and down-regulated genes in volcano plot.
            p_val: p value, to filter the differentially expressed genes for the volcano plot.
            y_lim_tuple: Values to adjust the Y-axis limits of the plot.
            save_plot: Boolean value to save plot.
            stdout : Flag to print logs to stdout.
        '''

        self.fixed_effect_column = fixed_effect_column
        self.fixed_effect_factors = fixed_effect_factors[::-1]
        self.group = group
        self.celltype_column = celltype_column
        self.cell_subsets = cell_subsets
        self.min_cell_threshold = min_cell_threshold
        self.n_cpu = n_cpu
        self.gene_batch_size = gene_batch_size
        self.coef_threshold = coef_threshold
        self.p_val = p_val
        self.y_lim_tuple = y_lim_tuple
        self.save_plot = save_plot
        self.stdout = stdout

        if self.n_cpu > multiprocessing.cpu_count():
            self.n_cpu = multiprocessing.cpu_count()

    def replace_spec_char_get_dict(self, var_names: pd.core.indexes.base.Index):
        ''' This method replaces any special character in gene names.

        Args:
            var_names: var_names in the Anndata.
            
        Returns:
            var_names with special characters replaced with '_', and a dictionary mapping of old and new names.
        '''

        old_new_name_map_dict = dict()
        special_chars = list(set(string.punctuation))

        def replace_special_chars(name):
            return ''.join(
                '_' if char in special_chars else char for char in name)

        replaced_name_array = np.vectorize(replace_special_chars)(var_names)
        for name in range(len(var_names)):
            old_new_name_map_dict[replaced_name_array[name]] = var_names[name]
        return replaced_name_array, old_new_name_map_dict

    def get_genes_n_fixed_val_subset_df(self,
                                        batch_adata: AnnData,
                                        cell_type: str = None):
        '''This method converts Anndata into a pandas DataFrame with gene expression data,
        'fixed_effect_column', and 'group' params.
        
        Args:
            batch_adata: Anndata.
            cell_type: Cell type in the 'celltype_column' to subset the anndata, 
                       the whole anndata will be processed if 'cell_type' is None.
            
        Returns:
            A list of gene names in the anndata, and a pandas dataframe with count matrix.
        '''

        if cell_type is not None:
            mask = pd.Series([True] * batch_adata.shape[0],
                             index=batch_adata.obs_names)
            for factor in self.fixed_effect_factors:
                assert factor in batch_adata.obs[
                    self.
                    fixed_effect_column].values, f"{factor} must be in the '{self.fixed_effect_column}' column in 'adata.obs'"
            mask &= batch_adata.obs[self.fixed_effect_column].isin(
                self.fixed_effect_factors)
            ad_subset = batch_adata[mask]
            #Remove unexpressed genes
            sc.pp.filter_genes(ad_subset, min_cells=self.min_cell_threshold)
            ad_subset_to_df = ad_subset.to_df()
            genes = ad_subset_to_df.columns.tolist()
            ad_subset_to_df[self.group] = ad_subset.obs[self.group].values
            ad_subset_to_df[self.fixed_effect_column] = ad_subset.obs[
                self.fixed_effect_column].values
        else:
            sc.pp.filter_genes(batch_adata, min_cells=self.min_cell_threshold)
            ad_subset_to_df = batch_adata.to_df()
            genes = ad_subset_to_df.columns.tolist()
            ad_subset_to_df[self.group] = batch_adata.obs[self.group].values
            ad_subset_to_df[self.fixed_effect_column] = batch_adata.obs[
                self.fixed_effect_column].values
        ad_subset_to_df[self.fixed_effect_column] = pd.Categorical(
            ad_subset_to_df[self.fixed_effect_column],
            categories=[*self.fixed_effect_factors],
            ordered=True)

        return genes, ad_subset_to_df

    def get_result_mxmodel_per_gene(self, gene: str,
                                    ad_subset_to_df: DataFrame):
        '''This method produces the Linear mixed-effects model statistics for a single gene.
        
        Args:
            gene: Gene name.
            ad_subset_to_df: A pandas dataframe with gene expression, 'fixed_effect_column', 
                             and 'group' params.
            
        Returns:
            A dictionary with model statistics.
        '''

        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)
                warnings.simplefilter("ignore", UserWarning)
                warnings.simplefilter("ignore", OptimizeWarning)
                warnings.simplefilter("ignore", ConvergenceWarning)
                warnings.simplefilter("ignore", HessianInversionWarning)
                formula = f'{gene} ~ {self.fixed_effect_column}'
                mxmodel = smf.mixedlm(formula,
                                      ad_subset_to_df,
                                      groups=ad_subset_to_df[self.group])
                mixmodel_result = mxmodel.fit()
                result_dict_per_gene = dict()
                result_dict_per_gene['gene'] = gene
                for category in self.fixed_effect_factors[1:]:
                    result_dict_per_gene[
                        f'coef_{category}'] = mixmodel_result.params[
                            f'{self.fixed_effect_column}[T.{category}]']
                    result_dict_per_gene[
                        f'SEcoef_{category}'] = mixmodel_result.bse[
                            f'{self.fixed_effect_column}[T.{category}]']
                    result_dict_per_gene[
                        f'pval_{category}'] = mixmodel_result.pvalues[
                            f'{self.fixed_effect_column}[T.{category}]']
                pickle.dumps(result_dict_per_gene)
                return result_dict_per_gene
        except LinAlgError as e:
            pass

    def get_multiproc_mxeffect_model_batch_res(self, gene_names: list[str],
                                               ad_subset_to_df: DataFrame):
        '''This method parallelizes the Linear mixed effects models for a list
        of genes.
        
        Args:
            gene_names: List of gene names.
            ad_subset_to_df: A pandas dataframe with gene expression, 'fixed_effect_column', 
            and 'group' params.
            
        Returns:
            A list of dictionaries with model stats for 'gene_names'.    
        '''

        mxmodel_results_list = Parallel(n_jobs=self.n_cpu, backend="loky")(
            delayed(self.get_result_mxmodel_per_gene)(gene, ad_subset_to_df)
            for gene in gene_names)
        mxmodel_results_list = [
            per_gene_result for per_gene_result in mxmodel_results_list
            if per_gene_result is not None
        ]
        return mxmodel_results_list

    def plot_lmem_dge_result(self,
                             lmem_res_df: DataFrame,
                             dirpath: str,
                             cell_type: str = None):
        '''This method produces a volcano plot for the model results for a data subset 
        with a cell type, or for the whole dataset.
        
        Args:
            lmem_res_df: A pandas DataFrame with Model results (p-value, co-efficients, Standard error..).
            dirpath: Path to save the plot.
            cell_type: Cell type used to subset input anndata.
        '''

        neg_log10_pval = -np.log10(self.p_val)
        for category in self.fixed_effect_factors:
            coef_col = next((col for col in lmem_res_df.columns
                             if col.startswith('coef') and category in col),
                            None)
            pval_col = next((col for col in lmem_res_df.columns
                             if col.startswith('pval') and category in col),
                            None)
            if (coef_col is not None) and (pval_col is not None):

                lmem_res_df[f'-log10_{pval_col}'] = -np.log10(
                    lmem_res_df[pval_col])
                down_reg_genes_idx = (lmem_res_df[coef_col] < (
                    self.coef_threshold)) & (lmem_res_df[f'-log10_{pval_col}']
                                             >= (neg_log10_pval))
                up_reg_genes_idx = (lmem_res_df[coef_col] > (
                    self.coef_threshold)) & (lmem_res_df[f'-log10_{pval_col}']
                                             >= (neg_log10_pval))
                rest_gene_idx = ~(down_reg_genes_idx | up_reg_genes_idx)
                plt.figure(figsize=(10, 5))
                plt.grid(False)
                plt.scatter(lmem_res_df.loc[up_reg_genes_idx, coef_col],
                            lmem_res_df.loc[up_reg_genes_idx,
                                            f'-log10_{pval_col}'],
                            color='red',
                            s=20,
                            alpha=0.25,
                            label=f"Upreg_genes(p<={self.p_val})")
                plt.scatter(lmem_res_df.loc[down_reg_genes_idx, coef_col],
                            lmem_res_df.loc[down_reg_genes_idx,
                                            f'-log10_{pval_col}'],
                            color='blue',
                            s=20,
                            alpha=0.25,
                            label=f"Downreg_genes(p<={self.p_val})")

                plt.scatter(lmem_res_df.loc[rest_gene_idx, coef_col],
                            lmem_res_df.loc[rest_gene_idx,
                                            f'-log10_{pval_col}'],
                            color='green',
                            alpha=0.2,
                            s=20,
                            label=f"Insignificant genes")
                if self.coef_threshold == 0:
                    plt.axvline(self.coef_threshold,
                                color='red',
                                linestyle='--',
                                alpha=0.4,
                                label=f'Coefficient({self.coef_threshold})')
                    plt.axhline(neg_log10_pval,
                                color='black',
                                linestyle='--',
                                alpha=0.4,
                                label=f'p_val ({self.p_val})')
                if self.coef_threshold != 0:
                    plt.axvline(self.coef_threshold,
                                color='red',
                                linestyle='--',
                                alpha=0.4,
                                label=f'Coefficient({self.coef_threshold})')
                    plt.axvline(-self.coef_threshold,
                                color='blue',
                                linestyle='--',
                                alpha=0.4,
                                label=f'Coefficient(-{self.coef_threshold})')
                    plt.axhline(neg_log10_pval,
                                color='black',
                                linestyle='--',
                                alpha=0.4,
                                label=f'p_val ({self.p_val})')
                plt.legend(loc='center left', bbox_to_anchor=(1, 0.6))

                plt.xlabel('Coefficient', fontweight='bold')
                plt.ylabel('-Log10(p-value)', fontweight='bold')
                _category = category.replace(' ', '')
                if cell_type is not None:
                    _cell_type = cell_type.replace(' ', '')
                    plt.title(
                        f'lmem_DGE of "{cell_type}" in "{category}" vs "{self.fixed_effect_factors[0]}"',
                        fontweight='bold')
                    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
                    plt.savefig(path.join(
                        dirpath, f'lmem_DGE_{_cell_type}_{_category}.svg'),
                                bbox_inches='tight')
                else:
                    plt.title(
                        f'lmem_DGE in "{category}" vs "{self.fixed_effect_factors[0]}"',
                        fontweight='bold')
                    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
                    plt.savefig(path.join(dirpath, f'lmemDGE_{_category}.svg'),
                                bbox_inches='tight')

    def generate_analysis(self, test_data: Union[AnnData, AnnCollection],
                          dirpath: str, **kwargs):
        '''This method calls methods to run multiple linear mixed effects models and  
        generate DGE results and plots.
        
        Args:
            test_data: Anndata.
            dirpath: Path to save results.
        '''

        if isinstance(test_data, AnnData):
            test_data = AnnCollection([test_data])
        logger = EventLogger('Differential Gene expression analysis',
                             stdout=self.stdout)
        logger.heading2("DGE analysis using LMEM")
        dirpath = os.path.join(dirpath, 'lmem_dge_result')
        os.makedirs(dirpath, exist_ok=True)
        new_var_names, varname_map_dict = self.replace_spec_char_get_dict(
            test_data.var_names)
        test_data.var_names = new_var_names
        result = False
        if self.celltype_column is not None:
            logger.info("Performing DGE analysis with subset anndata")
            fixed_val_list = list(test_data.obs[self.celltype_column].unique())
            if self.cell_subsets is not None:
                fixed_val_list = self.cell_subsets
            for cell_type in fixed_val_list:
                assert cell_type in test_data.obs[
                    self.
                    celltype_column].values, f"{cell_type} must be in the '{self.celltype_column}' column in 'adata.obs'"
                cell_type_test_data = test_data[test_data.obs[
                    self.celltype_column] == cell_type]
                logger.info(f'\nProcessing for "{cell_type}" ...')
                fixed_val_lmem_result_list = []
                for batch in range(0, len(cell_type_test_data.var_names),
                                   self.gene_batch_size):
                    with warnings.catch_warnings():
                        warnings.filterwarnings(
                            "ignore", category=ImplicitModificationWarning)
                        gene_list, batch_df = self.get_genes_n_fixed_val_subset_df(
                            cell_type_test_data[:, batch:batch +
                                                self.gene_batch_size].to_adata(
                                                ), cell_type)
                    result_lmem_batch = self.get_multiproc_mxeffect_model_batch_res(
                        gene_list, batch_df)
                    fixed_val_lmem_result_list.extend(result_lmem_batch)

                if fixed_val_lmem_result_list:
                    fixed_val_lmem_result_df = pd.DataFrame(
                        fixed_val_lmem_result_list)
                    standard_error_cols = [
                        col for col in fixed_val_lmem_result_df.columns
                        if col.startswith('SEcoef')
                    ]
                    pval_columns = [
                        col for col in fixed_val_lmem_result_df.columns
                        if col.startswith('pval')
                    ]
                    fixed_val_lmem_result_df.dropna(subset=pval_columns,
                                                    inplace=True)
                    for column in standard_error_cols:
                        condition_category = '_'.join(column.split('_')[1:])
                        fixed_val_lmem_result_df[
                            f'stat_{condition_category}'] = (
                                fixed_val_lmem_result_df[
                                    f'coef_{condition_category}'] /
                                (fixed_val_lmem_result_df[
                                    f'SEcoef_{condition_category}'].replace(
                                        0, np.nan)))
                    for column in pval_columns:
                        multitest_result_bh = multipletests(
                            fixed_val_lmem_result_df[column],
                            method='fdr_bh',
                            alpha=0.1)
                        fixed_val_lmem_result_df[
                            f'adj_{column}'] = multitest_result_bh[1]

                    for gene_name in fixed_val_lmem_result_df['gene']:
                        fixed_val_lmem_result_df[
                            'gene'] = fixed_val_lmem_result_df['gene'].replace(
                                {gene_name: varname_map_dict[gene_name]})
                    file_name = 'lmemDGE_' + (cell_type.replace(" ",
                                                                '')) + '.csv'
                    full_file_path = path.join(dirpath, file_name)
                    fixed_val_lmem_result_df.to_csv(full_file_path, index=False)
                    if self.save_plot:
                        plot = self.plot_lmem_dge_result(
                            fixed_val_lmem_result_df, dirpath, cell_type)
                        plt.close('all')
                    result = True
                else:
                    logger.info(f'Error in producing result for "{cell_type}"')
        else:
            logger.info("Performing DGE analysis with whole anndata ...")
            whole_data_lmem_result_list = []
            for batch in range(0, len(test_data.var_names),
                               self.gene_batch_size):
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        "ignore", category=ImplicitModificationWarning)
                    gene_list, batch_df = self.get_genes_n_fixed_val_subset_df(
                        test_data[:, batch:batch +
                                  self.gene_batch_size].to_adata())
                result_lmem_batch = self.get_multiproc_mxeffect_model_batch_res(
                    gene_list, batch_df)
                whole_data_lmem_result_list.extend(result_lmem_batch)

            if whole_data_lmem_result_list:
                whole_data_lmem_result_df = pd.DataFrame(
                    whole_data_lmem_result_list)
                standard_error_cols = [
                    col for col in whole_data_lmem_result_df.columns
                    if col.startswith('SEcoef')
                ]
                pval_columns = [
                    col for col in whole_data_lmem_result_df.columns
                    if col.startswith('pval')
                ]
                whole_data_lmem_result_df.dropna(subset=pval_columns,
                                                 inplace=True)
                for column in standard_error_cols:
                    condition_category = '_'.join(column.split('_')[1:])
                    whole_data_lmem_result_df[f'stat_{condition_category}'] = (
                        whole_data_lmem_result_df[f'coef_{condition_category}']
                        / (whole_data_lmem_result_df[
                            f'SEcoef_{condition_category}'].replace(0, np.nan)))
                for column in pval_columns:
                    multitest_result_bh = multipletests(
                        whole_data_lmem_result_df[column],
                        method='fdr_bh',
                        alpha=0.1)
                    whole_data_lmem_result_df[
                        f'adj_{column}'] = multitest_result_bh[1]

                for gene_name in whole_data_lmem_result_df['gene']:
                    whole_data_lmem_result_df[
                        'gene'] = whole_data_lmem_result_df['gene'].replace(
                            {gene_name: varname_map_dict[gene_name]})
                file_name = 'lmemDGE_whole' + '.csv'
                full_file_path = path.join(dirpath, file_name)
                whole_data_lmem_result_df.to_csv(full_file_path, index=False)
                if self.save_plot:
                    plot = self.plot_lmem_dge_result(whole_data_lmem_result_df,
                                                     dirpath)
                    plt.close('all')
                result = True
        if result:
            logger.info(f"\nLMEM-DGE results stored at: {dirpath}\n")

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for DgeLMEM_config."""
        return dict(fixed_effect_column='fixed_effect_column',
                    fixed_effect_factors='fixed_effect_factors',
                    group='group',
                    celltype_column=None,
                    cell_subsets=None,
                    min_cell_threshold=10,
                    n_cpu=6,
                    gene_batch_size=1000,
                    coef_threshold=0,
                    p_val=0.05,
                    y_lim_tuple=None,
                    save_plot=True,
                    stdout=False)



"""This file is an implementation of Tensorboard logging callback."""

import os
from os import path

import torch
from torch.utils.tensorboard import SummaryWriter

from scalr.nn.callbacks import CallbackBase


class TensorboardLogger(CallbackBase):
    """
    Tensorboard logging of the training process.

    Attributes:
        epoch: An interger count of epochs trained.
        writer: Object that writes to tensorboard.
    """

    def __init__(self, dirpath: str = '.'):
        """Intialize required parameters for tensorboard logging callback.

        Args:
            dirpath: Path of directory to store the experiment logs.
        """
        self.writer = SummaryWriter(path.join(dirpath, 'logs'))
        self.epoch = 0

    def __call__(self, train_loss: float, train_acc: float, val_loss: float,
                 val_acc: float, **kwargs):
        """Logs the train_loss, val_loss, train_accuracy, val_accuracy for each epoch."""
        self.epoch += 1
        self.writer.add_scalars('Loss', {
            'train': train_loss,
            'val': val_loss
        }, self.epoch)
        self.writer.add_scalars('Accuracy', {
            'train': train_acc,
            'val': val_acc
        }, self.epoch)

    @classmethod
    def get_default_params(cls):
        """Class method to get default params for model_config."""
        return dict(dirpath='.')



from ._callbacks import CallbackBase
from ._callbacks import CallbackExecutor
from .early_stopping import EarlyStopping
from .model_checkpoint import ModelCheckpoint
from .tensorboard_logger import TensorboardLogger



"""This is a test file for dge_pseudobulk.py"""

import os
from os import path
import shutil

import numpy as np

from scalr.analysis import dge_pseudobulk
from scalr.utils import generate_dummy_dge_anndata
from scalr.utils import read_data


def check_dge_result(result_path: str, expected_dge_result_dict: dict) -> None:
    """This function checks the expected DGE results with the generated results.

    Args:
        result_path: Path to the generated DGE results.
        expected_dge_result_dict: A dictionary with expected dge results.
    """

    celltype_list = list(expected_dge_result_dict.keys())
    result_files = os.listdir(result_path)
    csv_files = []
    svg_files = []
    for file in result_files:
        if file.endswith('.csv'):
            csv_files.append(file)
        elif file.endswith('.svg'):
            svg_files.append(file)

    # Checking for right numbers of csv & svg files.
    assert len(csv_files) == len(
        celltype_list), f"Expected {len(celltype_list)} csv files"
    assert len(svg_files) == len(
        celltype_list), f"Expected {len(celltype_list)} svg files"

    for celltype in celltype_list:
        celltype_csv = [file for file in csv_files if celltype in file]
        assert celltype_csv, f"CSV file for {celltype} is not produced"
        celltype_svg = [file for file in svg_files if celltype in file]
        assert celltype_svg, f"SVG file for {celltype} is not produced"

        celltype_df = read_data(path.join(result_path, celltype_csv[0]),
                                index_col=None)
        celltype_dge_result_dict = expected_dge_result_dict[celltype]
        assert celltype_df.shape == celltype_dge_result_dict['shape'], (
            f"There is a mismatch in the shape of the dge_result CSV file for '{celltype}'."
        )
        assert (
            celltype_df.loc[celltype_df.index[0],
                            'gene'] == celltype_dge_result_dict['gene'][0]
        ) & (np.round(
            celltype_df.loc[celltype_df.index[0],
                            celltype_dge_result_dict['random_col_and_val'][0]],
            2
        ) == celltype_dge_result_dict['random_col_and_val'][1]), (
            f'There is a mismatch in the DGE results of the first row in the CSV file for {celltype}'
        )
        assert (
            celltype_df.loc[celltype_df.index[-1],
                            'gene'] == celltype_dge_result_dict['gene'][-1]
        ) & (np.round(
            celltype_df.loc[celltype_df.index[-1],
                            celltype_dge_result_dict['random_col_and_val'][0]],
            2
        ) == celltype_dge_result_dict['random_col_and_val'][-1]), (
            f'There is a mismatch in the DGE results of the last row in the CSV file for {celltype}'
        )


# DgePseudoBulk parameters
pseudobulk_parms_dict = {
    'celltype_column': 'cell_type',
    'design_factor': 'disease',
    'factor_categories': ['disease_x', 'normal'],
    'sum_column': 'donor_id',
    'cell_subsets': ['B_cell', 'T_cell']
}
# Dictionary with expected results from DgePseudoBulk
expected_pbk_dge_result_dict = {
    'B_cell': {
        'shape': (10, 7),
        'gene': ['gene_1', 'gene_10'],
        'random_col_and_val': ('log2FoldChange', 0.97, 0.38)
    },
    'T_cell': {
        'shape': (10, 7),
        'gene': ['gene_1', 'gene_10'],
        'random_col_and_val': ('log2FoldChange', -0.26, 1.28)
    }
}


def test_pseudobulk_generate_analysis(
        dge_parms_dict: dict = pseudobulk_parms_dict,
        expected_dge_result_dict: dict = expected_pbk_dge_result_dict) -> None:
    """This function generates DGE result using `generate_analysis` method in DgePseudoBulk class.
    Finally checks the generated results with the expected by calling `check_dge_result`function.
    
    Args:
        dge_parms_dict: Parameters dictionary for the DgePseudoBulk class.
        expected_dge_result_dict: A dictionary with expected dge results.
    """

    os.makedirs('./tmp', exist_ok=True)

    # Generating dummy anndata.
    adata = generate_dummy_dge_anndata()

    # Path to store dge result.
    dirpath = './tmp'
    cell_subsets = list(expected_dge_result_dict.keys())
    dge_pbk = dge_pseudobulk.DgePseudoBulk(
        celltype_column=dge_parms_dict['celltype_column'],
        design_factor=dge_parms_dict['design_factor'],
        factor_categories=dge_parms_dict['factor_categories'],
        sum_column=dge_parms_dict['sum_column'],
        cell_subsets=cell_subsets)

    dge_pbk.generate_analysis(adata, dirpath)
    pbk_dirpath = path.join(dirpath, 'pseudobulk_dge_result')
    #Checking DGE result files and values for each celltype
    check_dge_result(pbk_dirpath, expected_dge_result_dict)

    shutil.rmtree('./tmp', ignore_errors=True)



"""This file is a base class for the model."""

from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
import torch
from torch import nn
from torch import Tensor
from torch.utils.data import DataLoader

import scalr
from scalr.utils import build_object


class ModelBase(nn.Module):
    """Class for the model.
    
    Contains different methods to make a forward() call, load, save weights
    and predict the data provided.
    """

    def __init__(self):
        super().__init__()

    def forward(self, x: Tensor) -> Tensor:
        """A function for forward pass of the model to generate outputs."""
        pass

    def load_weights(self, model_weights_path: str):
        """A function to initialize model weights from previous weights."""
        self.load_state_dict(torch.load(model_weights_path)['model_state_dict'])

    def save_weights(self, model_weights_path: str):
        """A function to save model weights at the path."""
        torch.save({'model_state_dict': self.state_dict()}, model_weights_path)

    def get_predictions(self, dl: DataLoader, device: str = 'cpu'):
        """A function to get predictions from the dataloader."""
        pass

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for model_config."""
        return dict()


def build_model(model_config: dict) -> tuple[nn.Module, dict]:
    """Builder object to get Model, updated model_config."""
    return build_object(scalr.nn.model, model_config)



from .data_utils import generate_dummy_anndata
from .data_utils import generate_dummy_dge_anndata
from .file_utils import load_full_data_from_config
from .file_utils import load_test_data_from_config
from .file_utils import load_train_val_data_from_config
from .file_utils import read_data
from .file_utils import write_chunkwise_data
from .file_utils import write_data
from .logger import EventLogger
from .logger import FlowLogger
from .misc_utils import build_object
from .misc_utils import overwrite_default
from .misc_utils import set_seed



from ._analyser import AnalysisBase
from ._analyser import build_analyser
from .dge_lmem import DgeLMEM
from .dge_pseudobulk import DgePseudoBulk
from .evaluation import generate_and_save_classification_report
from .evaluation import get_accuracy
from .gene_recall_curve import GeneRecallCurve
from .heatmap import Heatmap
from .roc_auc import RocAucCurve



"""This file is a class for data ingestion into the pipeline."""

from copy import deepcopy
import os
from os import path

import pandas as pd

from scalr.data.preprocess import build_preprocessor
from scalr.data.split import build_splitter
from scalr.utils import FlowLogger
from scalr.utils import read_data
from scalr.utils import write_data


class DataIngestionPipeline:
    """Class for Data Ingestion into the pipeline"""

    def __init__(self, data_config: dict, dirpath: str = '.'):
        """Load data config and create a `data` directory.

        Args:
            data_config (dict): Data processing configuration and paths.
            dirpath (str): Experiment data directory. Defaults to '.'.
        """

        self.flow_logger = FlowLogger('DataIngestion')

        self.data_config = deepcopy(data_config)
        self.target = self.data_config.get('target')
        self.sample_chunksize = self.data_config.get('sample_chunksize')
        self.num_workers = self.data_config.get('num_workers', 1)

        # Make some necessary checks and logs.
        if not self.target:
            self.flow_logger.warning('Target not given')

        if not self.sample_chunksize:
            self.flow_logger.warning(
                '''Sample chunk size not given. Will default to not using chunking.
                   Might results in excessive use of memory.''')

        self.datadir = dirpath

    def generate_train_val_test_split(self):
        """A function to split data into train, validation and test sets."""

        # TODO: Move to config validation
        if self.data_config['train_val_test'].get(
                'full_datapath',
                False) and self.data_config['train_val_test'].get(
                    'split_datapaths', False):
            self.flow_logger.warning(
                '''`full_datapath` and `split_datapaths` are both provided in
                the config. Pipeline will use `full_datapath` and it will overwrite
                the `split_datapaths`.
                ''')

        # TODO: Move to config validation
        if self.data_config['train_val_test'].get(
                'feature_subset_datapaths'
        ) or self.data_config['train_val_test'].get('final_datapaths'):
            raise ValueError(
                '''`final_datapaths` or `feature_subset_datapaths` can not be provided
                by the user in the config. These paths are generated by the pipeline!
                ''')

        if self.data_config['train_val_test'].get('full_datapath'):
            self.flow_logger.info('Generating Train, Validation and Test sets')
            if not self.target:
                self.flow_logger.warning(
                    '''Target not provided. Will not be able to perform
                    checks regarding splits.
                    ''')

            full_datapath = self.data_config['train_val_test']['full_datapath']
            self.full_data = read_data(full_datapath)
            splitter_config = deepcopy(
                self.data_config['train_val_test']['splitter_config'])
            splitter, splitter_config = build_splitter(splitter_config)
            self.data_config['train_val_test'][
                'splitter_config'] = splitter_config

            # Make data splits.
            train_val_test_split_indices = splitter.generate_train_val_test_split_indices(
                full_datapath, self.target)

            write_data(train_val_test_split_indices,
                       path.join(self.datadir, 'train_val_test_split.json'))

            # Check data splits.
            if self.target:
                splitter.check_splits(full_datapath,
                                      train_val_test_split_indices, self.target)

            # Write data splits.
            train_val_test_split_dirpath = path.join(self.datadir,
                                                     'train_val_test_split')
            os.makedirs(train_val_test_split_dirpath, exist_ok=True)

            splitter.write_splits(self.full_data, train_val_test_split_indices,
                                  self.sample_chunksize,
                                  train_val_test_split_dirpath,
                                  self.num_workers)

            # Garbage collection
            del self.full_data

            self.data_config['train_val_test'][
                'split_datapaths'] = train_val_test_split_dirpath

        elif self.data_config['train_val_test'].get('split_datapaths'):
            self.flow_logger.info(
                'Reading Train, Validation and Test sets from config')

        # TODO: Move to config validation
        else:
            raise ValueError(
                'No Data Provided. Please provide `full_datapath` or `split_datapaths`!'
            )

    def preprocess_data(self):
        """A function to apply preprocessing on data splits."""

        self.data_config['train_val_test']['final_datapaths'] = deepcopy(
            self.data_config['train_val_test']['split_datapaths'])

        all_preprocessings = self.data_config.get('preprocess', list())
        if not all_preprocessings:
            return

        self.flow_logger.info('Preprocessing data')
        datapath = self.data_config['train_val_test']['final_datapaths']

        processed_datapath = path.join(self.datadir, 'processed_data')
        os.makedirs(processed_datapath, exist_ok=True)

        for i, (preprocess) in enumerate(all_preprocessings):
            self.flow_logger.info(f'Applying {preprocess["name"]}')

            preprocessor, preprocessor_config = build_preprocessor(
                deepcopy(preprocess))
            self.data_config['preprocess'][i] = preprocessor_config
            # Fit on train data.
            preprocessor.fit(read_data(path.join(datapath, 'train')),
                             self.sample_chunksize)
            # Transform on train, val & test split.
            for split in ['train', 'val', 'test']:
                split_data = read_data(path.join(datapath, split))
                preprocessor.process_data(split_data, self.sample_chunksize,
                                          path.join(processed_datapath, split),
                                          self.num_workers)

            datapath = processed_datapath

        self.data_config['train_val_test'][
            'final_datapaths'] = processed_datapath

    def generate_mappings(self):
        """A function to generate an Integer mapping to and from target columns."""

        self.flow_logger.info(
            'Generate label mappings for all columns in metadata')

        column_names = read_data(
            path.join(self.data_config['train_val_test']['final_datapaths'],
                      'val')).obs.columns

        data_obs = []
        for split in ['train', 'val', 'test']:
            datapath = path.join(
                self.data_config['train_val_test']['final_datapaths'], split)
            split_data_obs = read_data(datapath).obs
            data_obs.append(split_data_obs)
        full_data_obs = pd.concat(data_obs)

        label_mappings = {}
        for column_name in column_names:
            label_mappings[column_name] = {}

            id2label = sorted(full_data_obs[column_name].astype(
                'category').cat.categories.tolist())

            label2id = {id2label[i]: i for i in range(len(id2label))}
            label_mappings[column_name]['id2label'] = id2label
            label_mappings[column_name]['label2id'] = label2id

        # Garbage collection
        del data_obs
        del full_data_obs

        write_data(label_mappings, path.join(self.datadir,
                                             'label_mappings.json'))

        self.data_config['label_mappings'] = path.join(self.datadir,
                                                       'label_mappings.json')

    def get_updated_config(self):
        """This function returns updated configs."""
        return self.data_config



"""This file generates differential gene expression analysis using Pseudobulk approach and stores the results."""

import os
from os import path
from typing import Optional, Tuple, Union

from anndata import AnnData
import anndata as ad
from anndata.experimental import AnnCollection
import matplotlib.pyplot as plt
import numpy as np
from pandas import DataFrame
import pandas as pd
from pydeseq2.dds import DeseqDataSet
from pydeseq2.ds import DeseqStats
import scanpy as sc

from scalr import utils
from scalr.analysis import AnalysisBase
from scalr.utils import EventLogger
from scalr.utils import read_data


class DgePseudoBulk(AnalysisBase):
    '''Class to perform differential gene expression analysis using Pseudobulk approach.'''

    def __init__(self,
                 celltype_column: str,
                 design_factor: str,
                 factor_categories: list[str],
                 sum_column: str,
                 cell_subsets: list[str] = None,
                 min_cell_threshold: int = 1,
                 fold_change: Union[float, int] = 1.5,
                 p_val: Union[float, int] = 0.05,
                 y_lim_tuple: Optional[Tuple[float, ...]] = None,
                 save_plot: bool = True,
                 stdout: bool = False):
        '''DgePseudoBulk parameters initialization.
        
        Args: 
            celltype_column: Column name in `anndata.obs` containing all the cell types.
            design_factor: Column name in `anndata.obs` containing different factor levels or categories for
                           differential gene expression analysis.         
            factor_categories: List of conditions in `design_factor` to make design matrix for.
            sum_column: Column name to sum values across samples.
            cell_subsets: Selcted list of cells in 'celltype_column' to subset the anndata.
            min_cell_threshold: Minimum number of subjects with aggregated nonzero gene expression values for a gene.
                                Used for filtering noisy genes.
            fold_change: Fold change to filter the differentially expressed genes for the volcano plot.
            p_val: p value, to filter the differentially expressed genes for the volcano plot.
            y_lim_tuple: Values to adjust the Y-axis limits of the plot.
            save_plot: Boolean value to save plot.
            stdout : Flag to print logs to stdout.
        '''

        self.celltype_column = celltype_column
        self.design_factor = design_factor
        # Hacky fix since pydeseq2 does not support `_` in `design_factor` or factor levels.
        self.design_factor_no_undrscr = design_factor.replace('_', '')
        self.factor_categories = [
            factor.replace('_', '-') for factor in factor_categories
        ]
        self.sum_column = sum_column
        self.cell_subsets = cell_subsets
        self.min_cell_threshold = min_cell_threshold
        self.fold_change = fold_change
        self.p_val = p_val
        self.y_lim_tuple = y_lim_tuple
        self.save_plot = save_plot
        self.stdout = stdout

    def _make_design_matrix(self, adata: AnnData, cell_type: str):
        '''Method to subset an anndata as per a cell type and 
        make design matrix based upon the factor levels in design_factor.
    
        Args:
            adata: AnnData.
            cell_type: Cell type to subset data on, belonging to `celltype_column`.
    
        Returns:
            AnnData oject of design matrix.
        '''

        if isinstance(adata, AnnData):
            adata = AnnCollection([adata])

        design_matrix_list = []

        fix_data = adata[adata.obs[self.celltype_column] == cell_type]
        for condition in self.factor_categories:
            condition_subset = fix_data[fix_data.obs[self.design_factor] ==
                                        condition]
            for sum_sample in condition_subset.obs[self.sum_column].unique():
                sum_subset = condition_subset[condition_subset.obs[
                    self.sum_column] == sum_sample]
                subdata = ad.AnnData(
                    X=sum_subset[:].X.sum(axis=0).reshape(
                        1, len(sum_subset.var_names)),
                    var=DataFrame(index=sum_subset.var_names),
                    obs=DataFrame(index=[f'{sum_sample}_{condition}']))
                subdata.obs[self.design_factor_no_undrscr] = [condition]
                design_matrix_list.append(subdata)

        design_matrix = ad.concat(design_matrix_list)
        return design_matrix

    def get_differential_expression_results(self, design_matrix: AnnData,
                                            cell_type: str, dirpath: str):
        '''Method to get differential gene expression analysis results.
    
        Args:
            design_matrix: AnnData generated using '_make_design_matrix'.
            cell_type: Cell type used to subset the input anndata.
            dirpath: Path to save the result.
    
        Returns:
            A pandas DataFrame object containing differential gene expression results.
        '''

        count_matrix = design_matrix.to_df()
        count_matrix_int = count_matrix.round().astype(int)
        metadata_ad = design_matrix.obs

        deseq_dataset = DeseqDataSet(
            counts=count_matrix_int,
            metadata=metadata_ad,
            design_factors=self.design_factor_no_undrscr)

        # removing genes with zero expression across all the sample
        sc.pp.filter_genes(deseq_dataset, min_cells=self.min_cell_threshold)
        deseq_dataset.deseq2()

        result_stats = DeseqStats(deseq_dataset,
                                  contrast=(self.design_factor_no_undrscr,
                                            *self.factor_categories))
        result_stats.summary()
        results_df = result_stats.results_df
        results_df.index.name = 'gene'

        if dirpath:
            _cell_type = cell_type.replace(" ", "")
            factor_0 = (self.factor_categories[0]).replace(" ", "")
            factor_1 = (self.factor_categories[1]).replace(" ", "")
            results_df.to_csv(
                path.join(dirpath,
                          f'pbkDGE_{_cell_type}_{factor_0}_vs_{factor_1}.csv'))
        return results_df

    def plot_volcano(self, dge_results_df: DataFrame, cell_type: str,
                     dirpath: str):
        '''Method to generate volcano plot of differential gene expression results 
        and store it on disk.
    
        Args:
            dge_results_df: Differential gene expression results in dataframe.
            cell_type: Cell type used to subset the input anndata.
            dirpath: Path to save the result.
        '''

        log2_fold_chnage = np.log2(self.fold_change)
        neg_log10_pval = -np.log10(self.p_val)
        dge_results_df['-log10(pvalue)'] = -np.log10(dge_results_df['pvalue'])

        upregulated_gene = (dge_results_df['log2FoldChange'] >= log2_fold_chnage
                           ) & (dge_results_df['-log10(pvalue)']
                                >= (neg_log10_pval))
        downregulated_gene = (dge_results_df['log2FoldChange'] <= (
            -log2_fold_chnage)) & (dge_results_df['-log10(pvalue)']
                                   >= (neg_log10_pval))

        unsignificant_gene = dge_results_df['-log10(pvalue)'] <= (
            neg_log10_pval)
        signi_genes_between_up_n_down = ~(upregulated_gene | downregulated_gene
                                          | unsignificant_gene)

        plt.figure(figsize=(10, 6))
        plt.scatter(dge_results_df.loc[upregulated_gene, 'log2FoldChange'],
                    dge_results_df.loc[upregulated_gene, '-log10(pvalue)'],
                    color='red',
                    alpha=0.2,
                    s=20,
                    label=f"FC>={self.fold_change} & p_val<={self.p_val}")

        plt.scatter(dge_results_df.loc[downregulated_gene, 'log2FoldChange'],
                    dge_results_df.loc[downregulated_gene, '-log10(pvalue)'],
                    color='blue',
                    alpha=0.2,
                    s=20,
                    label=f'FC<=-{self.fold_change} & p_val<={self.p_val}')

        plt.scatter(dge_results_df.loc[unsignificant_gene, 'log2FoldChange'],
                    dge_results_df.loc[unsignificant_gene, '-log10(pvalue)'],
                    color='mediumorchid',
                    alpha=0.2,
                    s=20,
                    label=f'p_val>{self.p_val}')

        plt.scatter(
            dge_results_df.loc[signi_genes_between_up_n_down, 'log2FoldChange'],
            dge_results_df.loc[signi_genes_between_up_n_down, '-log10(pvalue)'],
            color='green',
            alpha=0.2,
            s=20,
            label=
            f'-{self.fold_change}<FC<{self.fold_change} & p_val<={self.p_val}')

        plt.xlabel('Log2 Fold Change', fontweight='bold')
        plt.ylabel('-Log10(p-value)', fontweight='bold')
        plt.title(
            f'Psedobulk_DGE of "{cell_type}" in {self.factor_categories[0]} vs {self.factor_categories[1]}',
            fontweight='bold')
        plt.grid(False)
        plt.axhline(neg_log10_pval,
                    color='black',
                    linestyle='--',
                    label=f'p_val ({self.p_val})')
        plt.axvline(log2_fold_chnage,
                    color='red',
                    linestyle='--',
                    alpha=0.4,
                    label=f'Log2 Fold Change (+{self.fold_change})')
        plt.axvline(-log2_fold_chnage,
                    color='blue',
                    linestyle='--',
                    alpha=0.4,
                    label=f'Log2 Fold Change (-{self.fold_change})')
        plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        if self.y_lim_tuple:
            plt.ylim(bottom=self.y_lim_tuple[0], top=self.y_lim_tuple[1])
        _cell_type = cell_type.replace(" ", "")
        factor_0 = (self.factor_categories[0]).replace(" ", "")
        factor_1 = (self.factor_categories[1]).replace(" ", "")
        plt.savefig(path.join(
            dirpath, f'pbkDGE_{_cell_type}_{factor_0}_vs_{factor_1}.svg'),
                    bbox_inches='tight')

    def generate_analysis(self, test_data: Union[AnnData, AnnCollection],
                          dirpath: str, **kwargs):
        '''This method calls methods to perform differential gene expression analysis on data.
        
        Args:
            test_data: AnnData.
            dirpath: Path to save the result.
    
        Returns:
            Pandas DataFrame object containing differential gene expression stats.
        '''

        logger = EventLogger('Differential Gene expression analysis',
                             stdout=self.stdout)
        logger.heading2("DGE analysis using Pseudobulk")
        dirpath = os.path.join(dirpath, 'pseudobulk_dge_result')
        os.makedirs(dirpath, exist_ok=True)
        test_data.obs[self.design_factor] = test_data.obs[
            self.design_factor].str.replace('_', '-')
        assert self.celltype_column in test_data.obs.columns, f"{self.celltype_column} must be a column name in `adata.obs`"
        assert self.design_factor in test_data.obs.columns, f"{self.design_factor} must be a column name in `adata.obs`"
        assert self.sum_column in test_data.obs.columns, f"{self.sum_column} must be a column name in `adata.obs`"
        assert self.factor_categories[0] in test_data.obs[
            self.design_factor].unique(
            ), f"{self.factor_categories[0]} must belong to {self.design_factor}"
        assert self.factor_categories[1] in test_data.obs[
            self.design_factor].unique(
            ), f"{self.factor_categories[1]} must belong to '{self.design_factor}' column"

        cell_type_list = [
            condition
            for condition in test_data.obs[self.celltype_column].unique()
        ]

        if self.cell_subsets:
            cell_type_list = self.cell_subsets

        for cell_type in cell_type_list:
            assert cell_type in test_data.obs[self.celltype_column].unique(
            ), f"{cell_type} must belong to '{self.celltype_column}' column"
            logger.info(f'\nProcessing for "{cell_type}" ...')
            design_matrix = self._make_design_matrix(test_data, cell_type)
            dge_results_df = self.get_differential_expression_results(
                design_matrix, cell_type, dirpath)

            plot = self.plot_volcano(dge_results_df, cell_type, dirpath)
            plt.close(plot)
        logger.info(f"\nPseudobulk-DGE results stored at: {dirpath}\n")

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for DgePseudoBulk_config."""
        return dict(celltype_column='celltype_column',
                    design_factor='design_factor',
                    factor_categories='factor_categories',
                    sum_column='sum_column',
                    cell_subsets=None,
                    min_cell_threshold=1,
                    fold_change=1.5,
                    p_val=0.05,
                    y_lim_tuple=None,
                    save_plot=True,
                    stdout=False)



"""This is a test file for misc_utils.py"""

from copy import deepcopy

from scalr.utils import overwrite_default


def test_overwrite_default():
    """This funciton tests `overwrite_default()` function of misc_utils."""

    # User config key-values dictionary.
    user_config = {'a': 0, 'b': 1, 'd': 3}

    # Default config key-values dictionary.
    default_config = {'a': '5', 'b': 7, 'c': 2}

    # Getting updated default config using the overwrite function.
    updated_default_params = overwrite_default(
        user_config=user_config, default_config=deepcopy(default_config))

    # Checking whether the (key, values) not available in `user_config` are present in
    # `updated_default_params` and existing (key, values) are consistent with
    # `user_config` or not.
    for key in updated_default_params.keys():
        if key not in user_config:
            assert updated_default_params[key] == default_config[key]
        elif key in user_config:
            assert updated_default_params[key] == user_config[key]



'''This is a test file for simplemetadataloader.'''

import anndata
import numpy as np
import pandas as pd

from scalr.nn.dataloader import build_dataloader
from scalr.utils import generate_dummy_anndata


def test_metadataloader():

    # Generating dummy anndata.
    adata = generate_dummy_anndata(n_samples=15, n_features=7)

    # Generating mappings for anndata obs columns.
    mappings = {}
    for column_name in adata.obs.columns:
        mappings[column_name] = {}

        id2label = []
        id2label += adata.obs[column_name].astype(
            'category').cat.categories.tolist()

        label2id = {id2label[i]: i for i in range(len(id2label))}
        mappings[column_name]['id2label'] = id2label
        mappings[column_name]['label2id'] = label2id

    # Defining required parameters for metadataloader.
    metadata_col = ['batch', 'env']
    dataloader_config = {
        'name': 'SimpleMetaDataLoader',
        'params': {
            'batch_size': 10,
            'metadata_col': metadata_col
        }
    }
    dataloader, _ = build_dataloader(dataloader_config=dataloader_config,
                                     adata=adata,
                                     target='celltype',
                                     mappings=mappings)

    # Comparing expecting features shape after using metadatloader.
    for feature, _ in dataloader:
        assert feature.shape[1] == len(
            adata.var_names) + adata.obs[metadata_col].nunique().sum()
        # Breaking, as checking only first batch is enough.
        break



"""This file is a base class for implementation of Callbacks."""

import os
from os import path
from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection

import scalr
from scalr.utils import build_object


class CallbackBase:
    """Base class to build callbacks."""

    def __init__(self, dirpath='.'):
        """Use to generate necessary arguments or create directories."""
        pass

    def __call__(self):
        """Execute the callback here."""
        pass

    @classmethod
    def get_default_params(cls):
        """Class method to get default params for callbacks config."""
        return None


class CallbackExecutor:
    """
    Wrapper class to execute all enabled callbacks.

    Enabled callbacks are executed with the early stopping callback
    executed last to return a flag for continuation or stopping of model training
    """

    def __init__(self, dirpath: str, callbacks: list[dict]):
        """Intialize required parameters for callbacks.

        Args:
            dirpath: Path to store logs and checkpoints.
            callback: List containing multiple callbacks.
        """

        self.callbacks = []

        for callback in callbacks:
            if callback.get('params'):
                callback['params']['dirpath'] = dirpath
            else:
                callback['params'] = dict(dirpath=dirpath)
            callback_object, _ = build_object(scalr.nn.callbacks, callback)
            self.callbacks.append(callback_object)

    def execute(self, **kwargs) -> bool:
        """Execute all the enabled callbacks. Returns early stopping condition."""

        early_stop = False
        for callback in self.callbacks:
            # Below `| False` is to handle cases when callbacks return None.
            # And we want to return true when early stopping is achieved.
            early_stop |= callback(**kwargs) or False

        return early_stop



"""This file is a base class for feature scorer."""

from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
import numpy as np
from torch import nn

import scalr
from scalr.utils import build_object


class ScoringBase:
    """Base class for the scorer."""

    def __init__(self):
        pass

    # Abstract
    def generate_scores(self, model: nn.Module,
                        train_data: Union[AnnData, AnnCollection],
                        val_data: Union[AnnData, AnnCollection], target: str,
                        mappings: dict) -> np.ndarray:
        """A function to return the score of each feature for each class.

        Args:
            model (nn.Module): Trained model to generate scores from.
            train_data (Union[AnnData, AnnCollection]): Training data of model.
            val_data (Union[AnnData, AnnCollection]): Validation data of model.
            target (str): Column in data, used to train the model on.
            mappings (dict): Mapping of model output dimension to its
                             corresponding labels in the metadata columns.

        Returns:
            np.ndarray: score_matrix [num_classes X num_features]
        """
        pass

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params."""
        return dict()


def build_scorer(scorer_config: dict) -> tuple[ScoringBase, dict]:
    """Builder object to get scorer, updated scorer_config."""
    return build_object(scalr.feature.scoring, scorer_config)



'''This is a test file for standard-scaler normalization.'''

from sklearn import preprocessing

from scalr.data.preprocess import standard_scale
from scalr.utils import generate_dummy_anndata


def test_fit():
    '''This function tests fit() function of sample-norm normalization.
    
    fit() function is enough for testing, as we can compare mean and std with
    sklean standard-scaler object params.
    '''

    # Creating an annadata object.
    adata = generate_dummy_anndata(n_samples=100, n_features=25)

    # Standard scaler required parameters.
    with_mean = False
    with_std = True

    # scalr standard-scale normalization.
    scalr_std_scaler = standard_scale.StandardScaler(with_mean=with_mean,
                                                     with_std=with_std)
    # Required parameter - sample_chunksize to process data in chunks.
    sample_chunksize = 4
    scalr_std_scaler.fit(adata, sample_chunksize=sample_chunksize)

    # sklearn normalization
    sklearn_std_scaler = preprocessing.StandardScaler(with_mean=with_mean,
                                                      with_std=with_std)
    sklearn_std_scaler.fit(adata.X)

    # asserts to check the calculated mean and standard deviation, the error should be less than 1e-15.
    assert sum(
        abs(scalr_std_scaler.train_mean -
            sklearn_std_scaler.mean_).flatten() < 1e-15
    ) == adata.shape[1], "Train data mean is not correctly calculated..."
    assert sum(
        abs(scalr_std_scaler.train_std - sklearn_std_scaler.scale_).flatten() <
        1e-15) == adata.shape[
            1], "Train data standard deviation is not correctly calculated..."



"""This file contains functions related to file read-write."""

import json
from math import ceil
import os
from os import path
from typing import Union

from anndata import AnnData
import anndata as ad
from anndata.experimental import AnnCollection
from joblib import delayed
from joblib import Parallel
import numpy as np
import pandas as pd
import yaml

from scalr.utils.logger import FlowLogger


def read_data(
        filepath: str,
        backed: str = 'r',
        index_col: int = 0,
        return_anncollection: bool = True
) -> Union[dict, AnnData, AnnCollection]:
    """This function reads a json, yaml, csv or AnnData object file if the file path contains it.
    
    Returns an AnnCollection in case of a directory with chunked anndatas.

    Args:
        filepath (str): path to `json`, `yaml` or `h5ad` file.
                        Or directory containing multiple `h5ad` files.
        backed (str, optional): To load AnnData / AnnCollection in backed mode. Defaults to 'r'.

    Raises:
        ValueError: In case of the wrong file path provided.

    Returns:
        Union[dict, AnnData, AnnCollection].
    """
    if filepath.endswith('.json'):
        data = read_json(filepath)
    elif filepath.endswith('.yaml'):
        data = read_yaml(filepath)
    elif filepath.endswith('.csv'):
        data = read_csv(filepath, index_col=index_col)
    elif filepath.endswith('.h5ad'):
        data = read_anndata(filepath, backed=backed)
    elif path.exists(path.join(filepath, '0.h5ad')):
        data = read_chunked_anndatas(filepath,
                                     backed=backed,
                                     return_anncollection=return_anncollection)
    else:
        raise ValueError(
            '''`filepath` is not a `json`, `yaml`, `csv` or `h5ad` file,
            or directory containing `h5ad` files.
            ''')
    return data


def write_data(data: Union[dict, AnnData, pd.DataFrame], filepath: str):
    """This function writes data to `json`, `yaml`, `csv` or `h5ad` file."""
    if filepath.endswith('.json'):
        dump_json(data, filepath)
    elif filepath.endswith('.yaml'):
        dump_yaml(data, filepath)
    elif filepath.endswith('.csv'):
        dump_csv(data, filepath)
    elif filepath.endswith('.h5ad'):
        assert type(
            data
        ) == AnnData, 'Only AnnData objects can be written as `h5ad` files'
        dump_anndata(data, filepath)
    else:
        raise ValueError(
            '`filepath` does not contain `json`, `yaml`, or `h5ad` file')


def write_chunkwise_data(full_data: Union[AnnData, AnnCollection],
                         sample_chunksize: int,
                         dirpath: str,
                         sample_inds: list[int] = None,
                         feature_inds: list[int] = None,
                         transform: callable = None,
                         num_workers: int = 1):
    """This function writes data subsets iteratively in a chunkwise manner, to ensure
    only at most `sample_chunksize` samples are loaded at a time.

    This function can also apply transformation on each chunk.

    Args:
        full_data (Union[AnnData, AnnCollection]): data to be written in chunks.
        sample_chunksize (int): number of samples to be loaded at a time.
        dirpath (str): path/to/directory to write the chunks of data.
        sample_inds (list[int], optional): To be used in case of chunking
                                                       only a subset of samples.
                                                       Defaults to all samples.
        feature_inds (list[int], optional): To be used in case of writing
                                                        only a subset of features.dataframe.
                                                        Defaults to all features.
        transform (function): a function to apply a transformation on a chunked numpy array.
        num_workers (int): Number of jobs to run in parallel for data writing. Additional
                            workers will not use additional memory, but will be CPU-intensive.
    """
    if not path.exists(dirpath):
        os.makedirs(dirpath)

    if not num_workers:
        num_workers = 1

    if not sample_inds:
        sample_inds = list(range(len(full_data)))

    # Hacky fixes for an AnnCollection working/bug.
    if sample_chunksize >= len(sample_inds):
        sample_chunksize = len(sample_inds) - 1

    for i, (start) in enumerate(range(0, len(sample_inds), sample_chunksize)):
        if feature_inds:
            data = full_data[sample_inds[start:start + sample_chunksize],
                             feature_inds]
        else:
            data = full_data[sample_inds[start:start + sample_chunksize]]

        if not isinstance(data, AnnData):
            data = data.to_adata()
        data = data.to_memory(copy=True)

        for col in data.obs.columns:
            data.obs[col] = data.obs[col].astype('category')

        def transform_and_write_data(data: AnnData, chunk_number: int):
            """Internal function to transform a chunk of data and write 
            it to disk."""

            # Handling of empty data
            if len(data) == 0:
                return

            # Transformation
            if transform:
                data = AnnData(data.X, obs=data.obs, var=data.var)
                if not isinstance(data.X, np.ndarray):
                    data.X = data.X.A
                data.X = transform(data.X)

            write_data(data, path.join(dirpath, f'{chunk_number}.h5ad'))

        worker_chunksize = int(
            ceil(sample_chunksize /
                 num_workers)) if num_workers else sample_chunksize

        # Execute parallel jobs for transformation and witing of data.
        parallel = Parallel(n_jobs=num_workers)
        parallel(
            delayed(transform_and_write_data)(
                data=data[j * worker_chunksize:(j + 1) * worker_chunksize],
                chunk_number=i * num_workers + j) for j in range(num_workers))


def _get_datapath_from_config(data_config):
    """This function returns the datapath to be used to read from config.

    Args:
        data_config: Data config.
    """
    flow_logger = FlowLogger('File Utils')
    train_val_test_paths = data_config.get('train_val_test')
    datapath = None

    if not train_val_test_paths:
        raise ValueError('Split Datapaths not given')

    if train_val_test_paths.get('feature_subset_datapaths'):
        datapath = 'feature_subset_datapaths'
        flow_logger.info('Data Loaded from Feature subset datapaths')

    elif train_val_test_paths.get('final_datapaths'):
        datapath = 'final_datapaths'
        flow_logger.info('Data Loaded from Final datapaths')

    elif train_val_test_paths.get('split_datapaths'):
        datapath = 'split_datapaths'
        flow_logger.info('Data Loaded from Split datapaths')

    else:
        raise ValueError('Split Datapaths not given')

    return datapath


def load_train_val_data_from_config(data_config):
    """This function returns train & validation data from the data config.

    Args:
        data_config: Data config.
    """
    train_val_test_paths = data_config.get('train_val_test')
    datapath = _get_datapath_from_config(data_config)

    train_data = read_data(path.join(train_val_test_paths[datapath], 'train'))
    val_data = read_data(path.join(train_val_test_paths[datapath], 'val'))

    return train_data, val_data


def load_test_data_from_config(data_config):
    """This function returns test data from the data config.

    Args:
        data_config: Data config.
    """
    train_val_test_paths = data_config.get('train_val_test')
    datapath = _get_datapath_from_config(data_config)

    test_data = read_data(path.join(train_val_test_paths[datapath], 'test'))

    return test_data


def load_full_data_from_config(data_config):
    """This function returns full data from the data config.

    Args:
        data_config: Data config.
    """
    train_val_test_paths = data_config.get('train_val_test')
    datapath = _get_datapath_from_config(data_config)

    full_datas = []
    full_datas += read_data(path.join(train_val_test_paths[datapath], 'train'),
                            return_anncollection=False)
    full_datas += read_data(path.join(train_val_test_paths[datapath], 'val'),
                            return_anncollection=False)
    full_datas += read_data(path.join(train_val_test_paths[datapath], 'test'),
                            return_anncollection=False)

    full_data = AnnCollection(full_datas)
    return full_data


# Readers
def read_yaml(filepath: str) -> dict:
    """This function returns the config file loaded from yaml."""
    with open(filepath, 'r') as fh:
        config = yaml.safe_load(fh)
    return config


def read_json(filepath: str) -> dict:
    """This file returns the json file object."""
    with open(filepath, 'r') as fh:
        config = json.load(fh)
    return config


def read_csv(filepath: str, index_col: int = 0) -> pd.DataFrame:
    """This file returns the DataFrame file object."""
    return pd.read_csv(filepath, index_col=index_col)


def read_anndata(filepath: str, backed: str = 'r') -> AnnData:
    """This file returns the Anndata object from filepath."""
    data = ad.read_h5ad(filepath, backed=backed)
    return data


def read_chunked_anndatas(dirpath: str,
                          backed: str = 'r',
                          return_anncollection: bool = True) -> AnnCollection:
    """This file returns an AnnCollection object from multiple anndatas
    in dirpath directory.
    """
    datas = []
    for i in range(len(os.listdir(dirpath))):
        if os.path.exists(path.join(dirpath, f'{i}.h5ad')):
            datas.append(
                read_anndata(path.join(dirpath, f'{i}.h5ad'), backed=backed))
        else:
            break
    data = AnnCollection(datas)

    if return_anncollection:
        return data
    else:
        return datas


# Writers
def dump_json(config: dict, filepath: str):
    """This function stores the json file to filepath."""
    with open(filepath, 'w') as fh:
        config = json.dump(config, fh, indent=2)
    return


def dump_yaml(config: dict, filepath: str):
    """This function stores the config file to filepath."""
    with open(filepath, 'w') as fh:
        config = yaml.dump(config, fh)
    return


def dump_csv(df: pd.DataFrame, filepath: str):
    """This function stores the config file to filepath."""
    df.to_csv(filepath)
    return


def dump_anndata(adata: AnnData, filepath: str):
    """This function writes the AnnData to filepath."""
    adata.write(filepath, compression="gzip")



"""This file is an implementation of stratified group splitter."""

from pandas import DataFrame
from sklearn.model_selection import GroupShuffleSplit

from scalr.data.split import SplitterBase
from scalr.utils import read_data


class StratifiedGroupSplitter(SplitterBase):
    """Class for stratified group splitter.

    Generates split of data into train, validation, and test
    sets. Stratification ensures samples have the same value for `stratify`
    column, can not belong to different sets. Also, it ensures every split
    contains samples from each class available in the data.
    """

    def __init__(self, split_ratio: list[float], stratify: str):
        """Initialize splitter with required parameters.

        Args:
            split_ratio (list[float]): ratio to split number of samples in
            stratify (str): column name to metadata the split upon in `obs`
        """
        super().__init__()
        self.stratify = stratify
        self.split_ratio = split_ratio

    def _split_data_with_stratification(
            self, metadata: DataFrame, target: str,
            test_ratio: float) -> tuple[list[int], list[int]]:
        """A function to split given metadata into a training and testing set.

        Args:
            metadata (DataFrame): Dataframe containing all samples to be split.
            target (str): Target for classification present in `obs`.
            test_ratio (float): Ratio of samples belonging to the test split.

        Returns:
            (list(int), list(int)): Two lists consisting of train and test indices.
        """
        splitter = GroupShuffleSplit(test_size=test_ratio,
                                     n_splits=1,
                                     random_state=42)

        train_inds, test_inds = next(
            splitter.split(metadata,
                           metadata[target],
                           groups=metadata[self.stratify]))

        return train_inds, test_inds

    def generate_train_val_test_split_indices(self, datapath: str,
                                              target: str) -> dict:
        """A function to generate a list of indices for train/val/test split of the whole dataset.

        Args:
            datapath (str): Path to full data.
            target (str): Target for classification present in `obs`.

        Returns:
            dict: 'train', 'val' and 'test' indices list.
        """
        if not target:
            raise ValueError('Must provide target for StratifiedGroupSplitter')

        adata = read_data(datapath)
        metadata = adata.obs
        metadata['true_index'] = range(len(metadata))
        n_cls = metadata[target].nunique()

        if n_cls > 2:
            raise ValueError(
                'StratifiedGroupSplitter only works for binary classification.')

        total_ratio = sum(self.split_ratio)
        train_ratio = self.split_ratio[0] / total_ratio
        val_ratio = self.split_ratio[1] / total_ratio
        val_ratio = val_ratio / (val_ratio + train_ratio)
        test_ratio = self.split_ratio[2] / total_ratio

        train_indices = []
        val_indices = []
        test_indices = []

        for label in metadata[target].unique():
            label_metadata = metadata[metadata[target] == label]

            # Split testing and (train+val) indices.
            relative_train_val_inds, relative_test_inds = self._split_data_with_stratification(
                label_metadata, target, test_ratio)

            train_val_data = label_metadata.iloc[relative_train_val_inds]

            # Get train and val indices, relative to the `train_val_data`.
            relative_train_inds, relative_val_inds = self._split_data_with_stratification(
                train_val_data, target, val_ratio)

            # Get true_indices relative to the entire data.
            test_indices.extend(
                label_metadata.iloc[relative_test_inds]['true_index'].tolist())
            val_indices.extend(
                train_val_data.iloc[relative_val_inds]['true_index'].tolist())
            train_indices.extend(
                train_val_data.iloc[relative_train_inds]['true_index'].tolist())

        data_split = {
            'train': train_indices,
            'val': val_indices,
            'test': test_indices
        }

        return data_split

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for model_config."""
        return dict(split_ratio=[7, 1, 2], stratify='donor_id')



"""This file performs Sample-wise normalization on the data."""

from typing import Union

import numpy as np

from scalr.data.preprocess import PreprocessorBase


class SampleNorm(PreprocessorBase):
    """Class for Samplewise Normalization"""

    def __init__(self, scaling_factor: float = 1.0):
        """Initialize parameters for Sample-wise normalization.

        Args:
            scaling_factor: `Target sum` to maintain for each sample.
        """

        self.scaling_factor = scaling_factor

    def transform(self, data: np.ndarray) -> np.ndarray:
        """A function to transform provided input data.

        Args:
            data (np.ndarray): Input raw data.

        Returns:
            np.ndarray: Processed data.
        """
        data *= (self.scaling_factor / (data.sum(axis=1).reshape(len(data), 1)))
        return data

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for preprocess_config."""
        return dict(scaling_factor=1.0)



"""This file contains implementation for model training on feature subsets."""

from copy import deepcopy
import os
from os import path
from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
from joblib import delayed
from joblib import Parallel
from torch import nn

from scalr.model_training_pipeline import ModelTrainingPipeline
from scalr.utils import EventLogger
from scalr.utils import FlowLogger
from scalr.utils import read_data
from scalr.utils import write_chunkwise_data


class FeatureSubsetting:
    """Class for FeatureSubsetting.

    It trains a model for each subsetted datasets, each
    containing `feature_subsetsize` genes as features.
    """

    def __init__(self,
                 feature_subsetsize: int,
                 chunk_model_config: dict,
                 chunk_model_train_config: dict,
                 train_data: Union[AnnData, AnnCollection],
                 val_data: Union[AnnData, AnnCollection],
                 target: str,
                 mappings: dict,
                 dirpath: str = None,
                 device: str = 'cpu',
                 num_workers: int = 1,
                 sample_chunksize: int = None):
        """Initialize required parameters for feature subset training.

        Args:
            feature_subsetsize (int): Number of features in one subset.
            chunk_model_config (dict): Chunked model config.
            chunk_model_train_config (dict): Chunked model training config.
            train_data (Union[AnnData, AnnCollection]): Train dataset.
            val_data (Union[AnnData, AnnCollection]): Validation dataset.
            target (str): Target to train model.
            mappings (dict): mapping of target to labels.
            dirpath (str, optional): Dirpath to store chunked model weights. Defaults to None.
            device (str, optional): Device to train models on. Defaults to 'cpu'.
            num_workers (int, optional): Number of parallel processes to launch to train multiple
                                         feature subsets simultaneously. Defaults to using single
                                         process.
            sample_chunksize (int, optional): Chunks of samples to be loaded in memory at once. 
                                              Required when `num_workers` > 1.
        """
        self.feature_subsetsize = feature_subsetsize
        self.chunk_model_config = chunk_model_config
        self.chunk_model_train_config = chunk_model_train_config
        self.train_data = train_data
        self.val_data = val_data
        self.target = target
        self.mappings = mappings
        self.dirpath = dirpath
        self.device = device
        self.num_workers = num_workers if num_workers else 1
        self.sample_chunksize = sample_chunksize

        self.total_features = len(self.train_data.var_names)

        # Note that EventLogger does not work with parallel training
        # You may use tensorboard logging to track model training logs
        if self.num_workers == 1:
            self.event_logger = EventLogger('FeatureSubsetting')

    def write_feature_subsetted_data(self):
        """Write chunks of feature-subsetted data, to enable parallel training of models
        using different chunks of data."""
        if self.num_workers == 1:
            return

        self.feature_chunked_data_dirpath = path.join(self.dirpath,
                                                      'chunked_data')
        os.makedirs(self.feature_chunked_data_dirpath, exist_ok=True)

        i = 0
        for start in range(0, self.total_features, self.feature_subsetsize):

            feature_subset_inds = list(
                range(start,
                      min(start + self.feature_subsetsize,
                          self.total_features)))

            write_chunkwise_data(self.train_data,
                                 self.sample_chunksize,
                                 path.join(self.feature_chunked_data_dirpath,
                                           'train', str(i)),
                                 feature_inds=feature_subset_inds,
                                 num_workers=self.num_workers)

            write_chunkwise_data(self.val_data,
                                 self.sample_chunksize,
                                 path.join(self.feature_chunked_data_dirpath,
                                           'val', str(i)),
                                 feature_inds=feature_subset_inds,
                                 num_workers=self.num_workers)

            i += 1

        del self.train_data
        del self.val_data

    def train_chunked_models(self) -> list[nn.Module]:
        """Trains a model for each subset data.

        Returns:
            list[nn.Module]: List of models for each subset.
        """
        if self.num_workers == 1:
            self.event_logger.info('Feature subset models training')

        chunked_models_dirpath = path.join(self.dirpath, 'chunked_models')
        os.makedirs(chunked_models_dirpath, exist_ok=True)

        def train_chunked_model(i, start):
            if self.num_workers == 1:
                self.event_logger.info(f'\nChunk {i}')

            chunk_dirpath = path.join(chunked_models_dirpath, str(i))
            os.makedirs(chunk_dirpath, exist_ok=True)

            if self.num_workers > 1:
                train_features_subset = read_data(
                    path.join(self.feature_chunked_data_dirpath, 'train',
                              str(i)))
                val_features_subset = read_data(
                    path.join(self.feature_chunked_data_dirpath, 'val', str(i)))
            else:
                train_features_subset = self.train_data[:, start:start +
                                                        self.feature_subsetsize]
                val_features_subset = self.val_data[:, start:start +
                                                    self.feature_subsetsize]

            chunk_model_config = deepcopy(self.chunk_model_config)

            model_trainer = ModelTrainingPipeline(chunk_model_config,
                                                  self.chunk_model_train_config,
                                                  chunk_dirpath, self.device)

            model_trainer.set_data_and_targets(train_features_subset,
                                               val_features_subset, self.target,
                                               self.mappings)

            model_trainer.build_model_training_artifacts()
            best_model = model_trainer.train()

            self.chunk_model_config, self.chunk_model_train_config = model_trainer.get_updated_config(
            )

            return i, best_model

        parallel = Parallel(n_jobs=self.num_workers)
        models = parallel(
            delayed(train_chunked_model)(i, start) for i, (start) in enumerate(
                range(0, self.total_features, self.feature_subsetsize)))

        # parallel loop returns all models with the chunk number, which is used to sort models in order
        # model[1] returns only the model, without the chunk number
        models = sorted(models)
        models = [model[1] for model in models]
        return models

    def get_updated_configs(self):
        """Returns updated configs."""
        return self.chunk_model_config, self.chunk_model_train_config



"""This file is an implementation of a linear scorer."""

from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
import numpy as np
import torch
from torch import nn

from scalr.feature.scoring import ScoringBase


class LinearScorer(ScoringBase):
    """Class for the linear scorer.
    
    This Scorer is only applicable for linear (single-layer) models.
    It directly uses the weights as the score for each feature.
    """

    def __init__(self):
        pass

    def generate_scores(self, model: nn.Module, *args, **kwargs) -> np.ndarray:
        """A function to generate and return the weights of the model as a score."""
        return model.state_dict()['out_layer.weight'].cpu().detach().numpy()



"""This is a test file for early_stopping.py"""

from copy import deepcopy

from scalr.nn.callbacks import EarlyStopping


def test_early_stopping():
    """This function tests early stopping of the model."""

    # Custom-defined validation loss to check early stopping.
    val_losses = [5, 2, 3, 2.1, 1.9, 3.0, 2.5, 2.0, 0.7, 0.4]
    patience = 3

    # The model should early stop at epoch 8 (val_loss=2.0) based on defined patience.
    expected_early_stop_epoch = 8

    # Creating objects for early stopping.
    early_stop = EarlyStopping(patience=patience)

    # Iterating over above val_losses to test epoch at which it is early stopping.
    observed_epochs = 1
    for val_loss in val_losses:
        if early_stop.__call__(val_loss=deepcopy(val_loss)):
            break
        observed_epochs += 1

    assert observed_epochs==expected_early_stop_epoch, f"There is some issue in early stopping."\
    f" Expected epochs({expected_early_stop_epoch}) != observed epoch({observed_epochs}). Please check!"



"""This file is an implementation of a sequential model."""

from typing import Tuple

import torch
from torch import nn
from torch import Tensor
from torch.utils.data import DataLoader

from scalr.nn.model import ModelBase


class SequentialModel(ModelBase):
    """Class for Deep Neural Network model with linear layers."""

    def __init__(self,
                 layers: list[int],
                 dropout: float = 0,
                 activation: str = 'ReLU',
                 weights_init_zero: bool = False):
        """Initialize required parameters for the linear model.

        Args:
            layers (list[int]): List of layers' feature size going from
                                                 input_features to output_features.
            dropout (float, optional): Dropout after each layer.
                                       Floating point value [0,1).
                                       Defaults to 0.
            activation (str, optional): Activation function class after each layer.
                                        Defaults to 'ReLU'.
            weights_init_zero (bool, optional): [Bool] to initialize weights of the model to zero.
                                                Defaults to False.
        """
        super().__init__()

        try:
            activation = getattr(nn, activation)()
        except:
            raise ValueError(
                f'{activation} is not a valid activation function name in torch.nn'
            )

        dropout = nn.Dropout(dropout)

        self.layers = nn.ModuleList()
        n = len(layers)
        for i in range(n - 2):
            self.layers.append(nn.Linear(layers[i], layers[i + 1]))
            self.layers.append(activation)
            self.layers.append(dropout)
        self.out_layer = nn.Linear(layers[n - 2], layers[n - 1])

        self.weights_init_zero = weights_init_zero
        if weights_init_zero:
            self.make_weights_zero()

    def make_weights_zero(self):
        """A function to initialize layer weights to 0."""
        for layer in self.layers:
            torch.nn.init.constant_(layer.weight, 0.0)
        torch.nn.init.constant_(self.out_layer.weight, 0.0)

    def forward(self, x: Tensor) -> Tensor:
        """Pass input through the network.

            Args:
                x: Tensor, shape [batch_size, layers[0]].

            Returns:
                Output dict containing batched layers[-1]-dimensional vectors in ['cls_output'] key.
        """
        output = {}

        for i, layer in enumerate(self.layers):
            x = layer(x)
            output[f'layer{i}_output'] = x

        output['cls_output'] = self.out_layer(x)
        return output

    def get_predictions(
            self,
            dl: DataLoader,
            device: str = 'cpu'
    ) -> Tuple[list[int], list[int], list[list[int]]]:
        """A function to get predictions from a model, from the dataloader.

        Args:
            dl (DataLoader): DataLoader object containing samples.
            device (str, optional): Device to run the model on. Defaults to 'cpu'.

        Returns:
            True labels, Predicted labels, Predicted probabilities of all samples
            in the dataloader.
        """
        self.eval()
        test_labels, pred_labels, pred_probabilities = [], [], []

        for batch in dl:
            with torch.no_grad():
                x, y = batch[0].to(device), batch[1].to(device)
                out = self(x)['cls_output']

            test_labels += y.tolist()
            pred_labels += torch.argmax(out, dim=1).tolist()
            pred_probabilities += out.tolist()

        return test_labels, pred_labels, pred_probabilities

    @classmethod
    def get_default_params(cls):
        """Class method to get default params for model_config."""
        return dict(layers=None,
                    dropout=0,
                    activation='ReLU',
                    weights_init_zero=False)



"""This is a test file for dge_lmem.py"""

import os
from os import path
import shutil

import numpy as np

from scalr.analysis import dge_lmem
from scalr.analysis.test_dge_pseudobulk import check_dge_result
from scalr.utils import generate_dummy_dge_anndata
from scalr.utils import read_data

# DgeLMEM parameters
lmem_parms_dict = {
    'fixed_effect_column': 'disease',
    'fixed_effect_factors': ['disease_x', 'normal'],
    'group': 'donor_id',
    'celltype_column': 'cell_type'
}
# Dictionary with expected results from DgeLMEM
expected_lmem_dge_result_dict = {
    'B_cell': {
        'shape': (10, 6),
        'gene': ['gene_1', 'gene_10'],
        'random_col_and_val': ('coef_disease_x', 0.3, 0.02)
    },
    'T_cell': {
        'shape': (10, 6),
        'gene': ['gene_1', 'gene_10'],
        'random_col_and_val': ('coef_disease_x', -0.05, 0.37)
    }
}


def test_lmem_generate_analysis(
        dge_parms_dict: dict = lmem_parms_dict,
        expected_dge_result_dict: dict = expected_lmem_dge_result_dict) -> None:
    """This function generates DGE result using `generate_analysis` method in DgeLMEM class.
    Finally checks the generated results with the expected by calling `check_dge_result`function.
    
    Args:
        dge_parms_dict: Parameters dictionary for the DgeLMEM class.
        expected_dge_result_dict: A dictionary with expected dge results.
    """

    os.makedirs('./tmp', exist_ok=True)

    # Generating dummy anndata.
    adata = generate_dummy_dge_anndata()

    # Path to store dge result.
    dirpath = './tmp'
    cell_subsets = list(expected_dge_result_dict.keys())
    dge_lm = dge_lmem.DgeLMEM(
        fixed_effect_column=dge_parms_dict['fixed_effect_column'],
        fixed_effect_factors=dge_parms_dict['fixed_effect_factors'],
        group=dge_parms_dict['group'],
        celltype_column=dge_parms_dict['celltype_column'],
        cell_subsets=cell_subsets)

    dge_lm.generate_analysis(adata, dirpath)
    lmem_dirpath = path.join(dirpath, 'lmem_dge_result')
    #Checking DGE result files and values for each celltype
    check_dge_result(lmem_dirpath, expected_dge_result_dict)

    shutil.rmtree('./tmp', ignore_errors=True)



"""This file performs standard scaler normalization on the data."""

from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
import numpy as np

from scalr.data.preprocess import PreprocessorBase


class StandardScaler(PreprocessorBase):
    """Class for Standard Normalization"""

    def __init__(self, with_mean: bool = True, with_std: bool = True):
        """Initialize parameters for standard scaler normalization.

        Args:
            with_mean: Mean for standard scaling.
            with_std: Standard deviation for standard scaling.
        """

        self.with_mean = with_mean
        self.with_std = with_std

        # Parameters for standard scaler.
        self.train_mean = None
        self.train_std = None

    def transform(self, data: np.ndarray) -> np.ndarray:
        """A function to transform provided input data.

        Args:
            data (np.ndarray): raw data

        Returns:
            np.ndarray: processed data
        """
        if not self.with_mean:
            train_mean = np.zeros((1, data.shape[1]))
        else:
            train_mean = self.train_mean
        return (data - train_mean) / self.train_std

    def fit(self, data: Union[AnnData, AnnCollection],
            sample_chunksize: int) -> None:
        """This function calculate parameters for standard scaler object from the train data.
        
        Args:
            data: Data to calculate the required parameters of.
            sample_chunksize: Chunks of data that can be loaded into memory at once.
        
        """

        self.calculate_mean(data, sample_chunksize)
        self.calculate_std(data, sample_chunksize)

    def calculate_mean(self, data: Union[AnnData, AnnCollection],
                       sample_chunksize: int) -> None:
        """Function to calculate mean for each feature in the train data
        
        Args:
            data: Data to calculate the mean of.
            sample_chunksize: Chunks of data that can be loaded into memory at once.
            
        Returns:
            Nothing, stores mean per feature of the train data.
            """

        train_sum = np.zeros(data.shape[1]).reshape(1, -1)

        # Iterate through batches of data to get mean statistics
        for i in range(int(np.ceil(data.shape[0] / sample_chunksize))):
            train_sum += data[i * sample_chunksize:i * sample_chunksize +
                              sample_chunksize].X.sum(axis=0)
        self.train_mean = train_sum / data.shape[0]

    def calculate_std(self, data: Union[AnnData, AnnCollection],
                      sample_chunksize: int) -> None:
        """A function to calculate standard deviation for each feature in the train data.
        
        Args:
            data: Data to calculate the standard deviation of
            sample_chunksize: Chunks of data that can be loaded into memory at once.
            
        Returns:
            Nothing, stores standard deviation per feature of the train data.
            """

        # Getting standard deviation of entire train data per feature.
        if self.with_std:
            self.train_std = np.zeros(data.shape[1]).reshape(1, -1)
            # Iterate through batches of data to get std statistics
            for i in range(int(np.ceil(data.shape[0] / sample_chunksize))):
                self.train_std += np.sum(np.power(
                    data[i * sample_chunksize:i * sample_chunksize +
                         sample_chunksize].X - self.train_mean, 2),
                                         axis=0)
            self.train_std /= data.shape[0]
            self.train_std = np.sqrt(self.train_std)

            # Handling cases where standard deviation of feature is 0, replace it with 1.
            self.train_std[self.train_std == 0] = 1
        else:
            # If `with_std` is False, set train_std to 1.
            self.train_std = np.ones((1, data.shape[1]))

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for preprocess_config."""
        return dict(with_mean=True, with_std=True)



"""This file is a base class for splitter."""

import os
from os import path
from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection

import scalr
from scalr.utils import build_object
from scalr.utils import EventLogger
from scalr.utils import read_data
from scalr.utils import write_chunkwise_data
from scalr.utils import write_data


class SplitterBase:
    """Base class for splitter, to make Train|Val|Test Splits."""

    def __init__(self):
        self.event_logger = EventLogger('Splitter')

    # Abstract
    def generate_train_val_test_split_indices(datapath: str, target: str,
                                              **kwargs) -> dict:
        """Generate a list of indices for train/val/test split of whole dataset.

        Args:
            datapath (str): Path to full data.
            target (str): Target for classification present in `obs`.
            **kwargs: Any other params needed for splitting.

        Returns:
            dict: 'train', 'val' and 'test' indices list.
        """
        pass

    def check_splits(self, datapath: str, data_splits: dict, target: str):
        """This function performs certain checks regarding splits and logs
        the distribution of target classes in each split.

        Args:
            datapath (str): Path to full data.
            data_splits (dict): Split of 'train', 'val' and 'test' indices.
            target (str): Classification target column name in `obs`.
        """

        adata = read_data(datapath)
        metadata = adata.obs
        n_cls = metadata[target].nunique()

        train_inds = data_splits['train']
        val_inds = data_splits['val']
        test_inds = data_splits['test']

        # Check for classes present in splits.
        if len(metadata[target].iloc[train_inds].unique()) != n_cls:
            self.event_logger.warning(
                'All classes are not present in Train set')

        if len(metadata[target].iloc[val_inds].unique()) != n_cls:
            self.event_logger.warning(
                'All classes are not present in Validation set')

        if len(metadata[target].iloc[test_inds].unique()) != n_cls:
            self.event_logger.warning('All classes are not present in Test set')

        # Check for overlapping samples.
        assert len(set(train_inds).intersection(
            test_inds)) == 0, "Test and Train sets contain overlapping samples"
        assert len(
            set(val_inds).intersection(train_inds)
        ) == 0, "Validation and Train sets contain overlapping samples"
        assert len(set(test_inds).intersection(val_inds)
                  ) == 0, "Test and Validation sets contain overlapping samples"

        # LOGGING.
        self.event_logger.info('Train|Validation|Test Splits\n')
        self.event_logger.info(f'Length of train set: {len(train_inds)}')
        self.event_logger.info(f'Distribution of train set: ')
        self.event_logger.info(
            f'{metadata[target].iloc[train_inds].value_counts()}\n')

        self.event_logger.info(f'Length of val set: {len(val_inds)}')
        self.event_logger.info(f'Distribution of val set: ')
        self.event_logger.info(
            f'{metadata[target].iloc[val_inds].value_counts()}\n')

        self.event_logger.info(f'Length of test set: {len(test_inds)}')
        self.event_logger.info(f'Distribution of test set: ')
        self.event_logger.info(
            f'{metadata[target].iloc[test_inds].value_counts()}\n')

    def write_splits(self,
                     full_data: Union[AnnData, AnnCollection],
                     data_split_indices: dict,
                     sample_chunksize: int,
                     dirpath: int,
                     num_workers: int = None):
        """THis function writes the train validation and test splits to the disk.

        Args:
            full_data (Union[AnnData, AnnCollection]): Full data to be split.
            data_split_indices (dict): Indices of each split.
            sample_chunksize (int): Number of samples to be written in one file.
            dirpath (int): Path to write data into.
            num_workers (int): number of jobs to run in parallel for data writing.
        """

        for split in data_split_indices.keys():
            if sample_chunksize:
                split_dirpath = path.join(dirpath, split)
                os.makedirs(split_dirpath, exist_ok=True)
                write_chunkwise_data(full_data,
                                     sample_chunksize,
                                     split_dirpath,
                                     data_split_indices[split],
                                     num_workers=num_workers)
            else:
                filepath = path.join(dirpath, f'{split}.h5ad')
                write_data(full_data[data_split_indices[split]].to_memory(),
                           filepath)

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for model_config."""
        return dict()


def build_splitter(splitter_config: dict) -> tuple[SplitterBase, dict]:
    """Builder object to get splitter, updated splitter_config."""
    return build_object(scalr.data.split, splitter_config)



"""This file contains an implementation for the model training pipeline."""

from copy import deepcopy
import os
from os import path
from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
import torch

import scalr
from scalr.nn.callbacks import CallbackExecutor
from scalr.nn.dataloader import build_dataloader
from scalr.nn.loss import build_loss_fn
from scalr.nn.model import build_model
from scalr.utils import EventLogger
from scalr.utils import FlowLogger
from scalr.utils import load_train_val_data_from_config
from scalr.utils import read_data
from scalr.utils import set_seed
from scalr.utils import write_data


class ModelTrainingPipeline:
    """Class for Model training pipeline."""

    def __init__(self,
                 model_config: dict,
                 train_config: dict,
                 dirpath: str = None,
                 device: str = 'cpu'):
        """Initialize required parameters for model training pipeline.

        Class to get trained model from given configs

        Args:
            dirpath (str): Path to store checkpoints and logs of the model.
            model_config (dict): Model config.
            train_config (dict): Model training config.
            device (str, optional): Device to run model on. Defaults to 'cpu'.
        """
        self.flow_logger = FlowLogger('ModelTraining')
        set_seed(42)

        self.train_config = train_config
        self.model_config = model_config
        self.device = device
        self.dirpath = dirpath

    def load_data_and_targets_from_config(self, data_config: dict):
        """A function to load data and targets from data config.

        Args:
            data_config: Data config.
        """
        self.train_data, self.val_data = load_train_val_data_from_config(
            data_config)
        self.target = data_config.get('target')
        self.mappings = read_data(data_config['label_mappings'])

    def set_data_and_targets(self, train_data: Union[AnnData, AnnCollection],
                             val_data: Union[AnnData, AnnCollection],
                             target: Union[str, list[str]], mappings: dict):
        """A function to set data when you don't use data directly from config,
        but rather by other sources like feature subsetting, etc.

        Args:
            train_data (Union[AnnData, AnnCollection]): Training data.
            val_data (Union[AnnData, AnnCollection]): Validation data.
            target (Union[str, list[str]]): Target columns name(s).
            mappings (dict): Mapping of a column value to ids
                            eg. mappings[column_name][label2id] = {A: 1, B:2, ...}.
        """
        self.train_data = train_data
        self.val_data = val_data
        self.target = target
        self.mappings = mappings

    def build_model_training_artifacts(self):
        """This function configures the model, optimizer, and loss function required
        for model training.
        """
        self.flow_logger.info('Building model training artifacts')

        # Building model.
        self.model, self.model_config = build_model(self.model_config)
        self.model.to(self.device)

        # Building optimizer.
        opt_config = deepcopy(self.train_config.get('optimizer'))
        self.opt, opt_config = self.build_optimizer(
            self.train_config.get('optimizer'))
        self.train_config['optimizer'] = opt_config

        # Building Loss Function.
        self.loss_fn, loss_config = build_loss_fn(
            deepcopy(self.train_config.get('loss', dict())))
        self.train_config['loss'] = loss_config
        self.loss_fn.to(self.device)

        # Building Callbacks executor.
        self.callbacks = CallbackExecutor(
            self.dirpath, self.train_config.get('callbacks', list()))

        # Resuming from checkpoint using model weights.
        if self.train_config.get('resume_from_checkpoint'):
            self.flow_logger.info('Resuming model from checkpoint')
            self.flow_logger.info('Loading model weights...')
            self.model.load_weights(self.train_config['resume_from_checkpoint'])
            self.flow_logger.info('Loading optimizer state dict...')
            self.opt.load_state_dict(
                torch.load(self.train_config['resume_from_checkpoint'])
                ['optimizer_state_dict'])

    def build_optimizer(self, opt_config: dict = None):
        """A function to build optimizer.

        Args:
            opt_config (dict): Optimizer config.        
        """
        if not opt_config:
            opt_config = dict()
        name = opt_config.get('name', 'Adam')
        opt_config['name'] = name
        params = opt_config.get('params', dict(lr=1e-3))
        opt_config['params'] = params

        opt = getattr(torch.optim, name)(self.model.parameters(), **params)
        return opt, opt_config

    def train(self):
        """This function trains the model."""
        self.flow_logger.info('Training the model')
        # Building Trainer.
        trainer_name = self.train_config.get('trainer', 'SimpleModelTrainer')
        self.train_config['trainer'] = trainer_name

        Trainer = getattr(scalr.nn.trainer, trainer_name)
        trainer = Trainer(self.model, self.opt, self.loss_fn, self.callbacks,
                          self.device)

        # Building DataLoaders.
        dataloader_config = self.train_config.get('dataloader')
        train_dl, dataloader_config = build_dataloader(dataloader_config,
                                                       self.train_data,
                                                       self.target,
                                                       self.mappings)
        val_dl, dataloader_config = build_dataloader(dataloader_config,
                                                     self.val_data, self.target,
                                                     self.mappings)
        self.train_config['dataloader'] = dataloader_config

        epochs = self.train_config.get('epochs', 1)
        self.train_config['epochs'] = epochs

        # Train and store the best model.
        best_model = trainer.train(epochs, train_dl, val_dl)
        if self.dirpath:
            best_model_dir = path.join(self.dirpath, 'best_model')
            os.makedirs(best_model_dir, exist_ok=True)
            best_model.save_weights(path.join(best_model_dir, 'model.pt'))
            write_data(self.model_config,
                       path.join(best_model_dir, 'model_config.yaml'))
            write_data(self.mappings, path.join(best_model_dir,
                                                'mappings.json'))

        return best_model

    def get_updated_config(self):
        """This function returns updated configs."""
        return self.model_config, self.train_config



"""This file is an implementation of SHAP scorer."""

from typing import Tuple, Union

from anndata import AnnData
from anndata.experimental import AnnCollection
import numpy as np
import pandas as pd
import shap
from sklearn.preprocessing import OneHotEncoder
import torch
from torch import nn

from scalr import utils
from scalr.feature.scoring import ScoringBase
from scalr.nn.dataloader import build_dataloader
from scalr.nn.model import CustomShapModel


class ShapScorer(ScoringBase):
    """Class for SHAP scorer. It can be used for any model."""

    def __init__(self,
                 early_stop: dict,
                 dataloader: dict,
                 device: str = 'cpu',
                 top_n_genes: int = 100,
                 background_tensor: int = 200,
                 samples_abs_mean: bool = True,
                 logger: str = 'EventLogger',
                 *args,
                 **kwargs):
        """Initialize class with SHAP arguments.

        Args:
            early_stop: Contains early stopping-related configuration.
            dataloader: Dataloader related config.
            device: Where data is processed/loaded.
            top_n_genes: Top N genes for each class/label.
            background_tensor: Number of training data used for SHAP explainer.
            samples_abs_mean: Apply abs before taking the mean across samples.
        """

        self.early_stop_config = early_stop
        self.device = device
        self.top_n_genes = top_n_genes
        self.background_tensor = background_tensor
        self.dataloader_config = dataloader
        self.samples_abs_mean = samples_abs_mean

        self.logger = getattr(utils, logger)('SHAP analysis')

    def generate_scores(self, model: nn.Module,
                        train_data: Union[AnnData, AnnCollection],
                        val_data: Union[AnnData, AnnCollection], target: str,
                        mappings: dict, *args, **kwargs) -> np.ndarray:
        """This function returns the weights of the model as a score.

        Args:
            model: Trained model that is used for SHAP.
            train_data: Data that is used as reference data for SHAP.
            val_data: On which SHAP will generate the score.
            mappings: Contains target-related mappings.

        Returns:
            class * genes abs weights matrix.
        """

        shap_values = self.get_top_n_genes_weights(model, train_data, val_data,
                                                   target, mappings)

        return shap_values

    def get_top_n_genes_weights(
            self, model: nn.Module, train_data: Union[AnnData, AnnCollection],
            test_data: Union[AnnData, AnnCollection], target: str,
            mappings: dict) -> Tuple[np.ndarray, np.ndarray]:
        """ A function to get top n genes of each class and its weights.

        Args:
            model: Trained model to extract weights from.
            train_data: Train data.
            test_data: Test data that is used for SHAP values.
            target: Target name.
            mappings: Contains target-related mappings.

        Returns:
            (class * genes abs weights matrix, class * genes weights matrix).
        """

        if isinstance(self.logger, utils.EventLogger):
            self.logger.heading2("Genes analysis using SHAP.")

        model.to(self.device)
        shap_model = CustomShapModel(model)

        random_indices = np.random.randint(0, train_data.shape[0],
                                           self.background_tensor)
        train_dl, _ = build_dataloader(self.dataloader_config,
                                       train_data[random_indices], target,
                                       mappings)
        random_background_data = torch.cat([batch[0] for batch in train_dl])

        self.logger.info(
            f"Selected random background data: {random_background_data.shape}")

        test_dl, _ = build_dataloader(self.dataloader_config, test_data, target,
                                      mappings)

        explainer = shap.DeepExplainer(shap_model,
                                       random_background_data.to(self.device))

        prev_top_genes_batch_wise = {}
        count_patience = 0
        total_samples = 0

        for batch_id, batch in enumerate(test_dl):
            self.logger.info(f"Running on batch: {batch_id}")
            total_samples += batch[0].shape[0]

            batch_shap_values = explainer.shap_values(batch[0].to(self.device))
            if self.samples_abs_mean:
                sum_shap_values = np.abs(batch_shap_values).sum(axis=0)
            else:
                # Calcluating 2 mean with abs values and non-abs values.
                # Non-abs values required for heatmap.
                sum_shap_values = batch_shap_values.sum(axis=0)

            if batch_id >= 1:
                sum_shap_values = np.sum(
                    [sum_shap_values, prev_batches_sum_shap_values], axis=0)

            mean_shap_values = sum_shap_values / total_samples

            genes_class_shap_df = pd.DataFrame(
                mean_shap_values[:len(test_dl.dataset.var_names)],
                index=test_dl.dataset.var_names)

            prev_batches_sum_shap_values = sum_shap_values

            early_stop, prev_top_genes_batch_wise = self._is_shap_early_stop(
                batch_id, genes_class_shap_df, prev_top_genes_batch_wise,
                self.top_n_genes, self.early_stop_config['threshold'])

            count_patience = count_patience + 1 if early_stop else 0

            if count_patience == self.early_stop_config['patience']:
                self.logger.info(f"Early stopping at batch: {batch_id}")
                break

        return mean_shap_values.T

    def _is_shap_early_stop(
        self,
        batch_id: int,
        genes_class_shap_df: pd.DataFrame,
        prev_top_genes_batch_wise: dict,
        top_n_genes: int,
        threshold: int,
    ) -> Tuple[bool, dict]:
        """A function to check whether previous and current batches' common genes are
            are greater than or equal to the threshold and return top genes
            batch wise.

        Args:
            batch_id: Current batch number.
            genes_class_shap_df: label/class wise genes SHAP values(mean across samples).
            prev_top_genes_batch_wise: Dictionary where prev batches per labels top genes are stored.
            top_n_genes: Number of top genes check.
            threshold: early stop if common genes are higher than this.

        Returns:
            Early stop value, top genes batch wise.
        """

        early_stop = True
        top_genes_batch_wise = {}
        classes = genes_class_shap_df.columns
        for label in classes:
            top_genes_batch_wise[label] = genes_class_shap_df[
                label].sort_values(ascending=False)[:top_n_genes].index

            # Start checking after first batch.
            if batch_id >= 1:
                num_common_genes = len(
                    set(top_genes_batch_wise[label]).intersection(
                        set(prev_top_genes_batch_wise[label])))
                # If commnon genes are less than 90 early stop will be false.
                if num_common_genes < threshold:
                    early_stop = False
            else:
                early_stop = False

        return early_stop, top_genes_batch_wise

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params."""
        return {
            "top_n_genes": 100,
            "background_tensor": 200,
            "samples_abs_mean": True,
            "early_stop": {
                "patience": 5,
                "threshold": 95
            },
            "dataloader": {
                "name": "SimpleDataLoader",
                "params": {
                    "batch_size": 5000,
                    "padding": 5000
                }
            }
        }



"""This file is an implementation of the stratified splitter."""

from pandas import DataFrame
from sklearn.model_selection import StratifiedShuffleSplit

from scalr.data.split import SplitterBase
from scalr.utils import read_data


class StratifiedSplitter(SplitterBase):
    """ Generate Stratified split of data into train, validation, and test sets.

    Stratification ensures the percentage of samples for each class. It ensures
    every split contains samples from each class available in the data.
    """

    def __init__(self, split_ratio: list[float]):
        """Initialize splitter with required parameters.

        Args:
            split_ratio (list[float]): ratio to split number of samples in
        """
        super().__init__()
        self.split_ratio = split_ratio

    def _split_data_with_stratification(
            self, metadata: DataFrame, target: str,
            test_ratio: float) -> tuple[list[int], list[int]]:
        """A function to split the given metadata into a training and testing set.

        Args:
            metadata (DataFrame): Dataframe containing all samples to be split.
            target (str): Target for classification present in `obs`.
            test_ratio (float): Ratio of samples belonging to the test split.

        Returns:
            (list(int), list(int)): Two lists consisting of train and test indices.
        """
        splitter = StratifiedShuffleSplit(test_size=test_ratio,
                                          n_splits=1,
                                          random_state=42)

        train_inds, test_inds = next(splitter.split(metadata, metadata[target]))

        return train_inds, test_inds

    def generate_train_val_test_split_indices(self, datapath: str,
                                              target: str) -> dict:
        """A function to generate a list of indices for train/val/test split of the whole dataset.

        Args:
            datapath (str): Path to full data.
            target (str): Target for classification present in `obs`.

        Returns:
            dict: 'train', 'val' and 'test' indices list.
        """
        if not target:
            raise ValueError('Must provide target for StratifiedSplitter')

        adata = read_data(datapath)
        metadata = adata.obs
        metadata['true_index'] = range(len(metadata))
        n_cls = metadata[target].nunique()

        total_ratio = sum(self.split_ratio)
        train_ratio = self.split_ratio[0] / total_ratio
        val_ratio = self.split_ratio[1] / total_ratio
        test_ratio = self.split_ratio[2] / total_ratio

        # Split testing and (train+val) indices.
        training_inds, testing_inds = self._split_data_with_stratification(
            metadata, target, test_ratio)

        train_val_data = metadata.iloc[training_inds]
        val_ratio = val_ratio / (val_ratio + train_ratio)

        # Get train and val indices, relative to the `train_val_data`.
        relative_train_inds, relative_val_inds = self._split_data_with_stratification(
            train_val_data, target, val_ratio)

        # Get true_indices relative to the entire data.
        true_test_inds = testing_inds.tolist()
        true_val_inds = train_val_data.iloc[relative_val_inds][
            'true_index'].tolist()
        true_train_inds = train_val_data.iloc[relative_train_inds][
            'true_index'].tolist()

        data_split = {
            'train': true_train_inds,
            'val': true_val_inds,
            'test': true_test_inds
        }

        return data_split

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for model_config."""
        return dict(split_ratio=[7, 1, 2])



"""This file is an implementation of the custom SHAP model."""

from torch import nn


class CustomShapModel(nn.Module):
    """Class for a custom model for SHAP."""

    def __init__(self, model, key='cls_output'):
        """Initialize required parameters for SHAP model.

        Args:
            model: Trained model used for SHAP calculation.
            key: key from model output dict.
        """
        super().__init__()
        self.model = model
        self.key = key

    def forward(self, x):
        """Pass input through the model and return output.
        
        Args:
            x: Tensor.
        """
        output = self.model(x)

        if isinstance(output, dict):
            output = output[self.key]

        return output



"""This is a test file for simpledataloader."""

from scalr.nn.dataloader import build_dataloader
from scalr.utils import generate_dummy_anndata


def test_metadataloader():
    """This function tests features shape returned by simpledataloader for the below 2 cases.
        1. #features are consistent with feature_subsetsize. No padding is required.
        2. #features are less than feature_subsetsize. This case needs padding.
    """

    # Generating dummy anndata.
    n_samples = 30
    n_features = 13
    adata = generate_dummy_anndata(n_samples=n_samples, n_features=n_features)

    # Generating mappings for anndata obs columns.
    mappings = {}
    for column_name in adata.obs.columns:
        mappings[column_name] = {}

        id2label = []
        id2label += adata.obs[column_name].astype(
            'category').cat.categories.tolist()

        label2id = {id2label[i]: i for i in range(len(id2label))}
        mappings[column_name]['id2label'] = id2label
        mappings[column_name]['label2id'] = label2id

    #                         Test case 1
    # Expected features shape after dataloading is (batch_size, 13).
    # So no padding is required as adata n_features=13. But we can pass
    # `padding=feature_subsetsize` in dataloader_config.

    ## Defining required parameters for simpledataloader.
    feature_subsetsize = 13
    dataloader_config = {
        'name': 'SimpleDataLoader',
        'params': {
            'batch_size': 10,
            'padding': feature_subsetsize,
        }
    }
    dataloader, _ = build_dataloader(dataloader_config=dataloader_config,
                                     adata=adata,
                                     target='celltype',
                                     mappings=mappings)

    ## Comparing expecting features shape after using metadatloader.
    for feature, _ in dataloader:
        assert feature.shape[
            1] == feature_subsetsize, f"There is some issue in simpledataloader."\
        f" Expected #features({n_features}) != Observed #features({feature.shape[1]}). Please check!"
        # Breaking, as checking only the first batch is enough.
        break

    #                         Test case 2
    # Expected features shape after dataloading is (batch_size, 20).
    # So padding is required as adata n_features=13. Hence 7 columns should be padded in dataloader.

    ## Defining required parameters for simpledataloader.
    feature_subsetsize = 20
    dataloader_config = {
        'name': 'SimpleDataLoader',
        'params': {
            'batch_size': 10,
            'padding': feature_subsetsize,
        }
    }
    dataloader, _ = build_dataloader(dataloader_config=dataloader_config,
                                     adata=adata,
                                     target='celltype',
                                     mappings=mappings)

    ## Comparing expected features shape after using metadatloader.
    for feature, _ in dataloader:
        assert feature.shape[
            1] == feature_subsetsize, f"There is some issue in simpledataloader."\
        f" Expected #features({feature_subsetsize}) != Observed #features({feature.shape[1]}). Please check!"
        # Breaking, as checking only the first batch is enough.
        break



"""This file is a base class for preprocessing module."""

from os import path
from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
import numpy as np

import scalr
from scalr.utils import build_object
from scalr.utils import write_chunkwise_data


class PreprocessorBase:
    """Base class for preprocessor"""

    def __init__(self, **kwargs):
        # Store all params here.
        pass

    # Abstract
    def transform(self, data: np.ndarray) -> np.ndarray:
        """A required function to transform a numpy array.

        Args:
            data (np.ndarray): Input raw data.

        Returns:
            np.ndarray: Processed data.
        """
        pass

    def fit(
        self,
        data: Union[AnnData, AnnCollection],
        sample_chunksize: int,
    ) -> None:
        """A function to calculate attributes for transformation.
        
        It is required only when you need to see the entire train data and
        calculate attributes, as required in StdScaler, etc. This method
        should not return anything, it should be used to store attributes
        that will be used by the `transform` method.

        Args:
            data (Union[AnnData, AnnCollection]): train_data in backed mode.
            sample_chunksize (int): several samples of data that can at most
                                    be loaded in memory.
        """
        pass

    def process_data(self,
                     full_data: Union[AnnData, AnnCollection],
                     sample_chunksize: int,
                     dirpath: str,
                     num_workers: int = 1):
        """A function to process the entire data chunkwise and write the processed data
        to disk.

        Args:
            full_data (Union[AnnData, AnnCollection]): Full data for transformation.
            sample_chunksize (int): Number of samples in one chunk.
            dirpath (str): Path to write the data to.
            num_workers (int): number of jobs to run in parallel for data writing.
        """
        if not sample_chunksize:
            # TODO
            raise NotImplementedError(
                'Preprocessing does not work without sample chunk size')

        write_chunkwise_data(full_data,
                             sample_chunksize,
                             dirpath,
                             transform=self.transform,
                             num_workers=num_workers)


def build_preprocessor(
        preprocessing_config: dict) -> tuple[PreprocessorBase, dict]:
    """Builder object to get a processor, updated preprocessing_config."""
    return build_object(scalr.data.preprocess, preprocessing_config)



"""This file contains an implementation of model evaluation and performs downstream analysis tasks."""

from copy import deepcopy
import os
from os import path
from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection

from scalr.analysis import build_analyser
from scalr.analysis.evaluation import generate_and_save_classification_report
from scalr.analysis.evaluation import get_accuracy
from scalr.feature_extraction_pipeline import FeatureExtractionPipeline
from scalr.nn.dataloader import build_dataloader
from scalr.nn.model import build_model
from scalr.utils import FlowLogger
from scalr.utils import load_full_data_from_config
from scalr.utils import load_test_data_from_config
from scalr.utils import load_train_val_data_from_config
from scalr.utils import read_data


class EvalAndAnalysisPipeline:
    """Class for evaluation and analysis of the trained model."""

    def __init__(self, analysis_config, dirpath, device):
        """Initialize required parameters for analysis.
        
        Args:
            analysis_config: Analysis config.
            dirpath: Path to store analysis outputs.
        """
        self.flow_logger = FlowLogger('Eval&Analysis')

        self.analysis_config = deepcopy(analysis_config)
        self.dirpath = dirpath
        self.device = device

        model_checkpoint = self.analysis_config.get('model_checkpoint')
        if model_checkpoint:
            model_config = read_data(
                path.join(model_checkpoint, 'model_config.yaml'))
            model_weights = path.join(model_checkpoint, 'model.pt')
            self.model, _ = build_model(model_config)
            self.model.to(self.device)
            self.model.load_weights(model_weights)
        else:
            self.flow_logger.warning(
                'Model path not provided. Unable to perform model-based analysis!'
            )
            self.model = None

        # Dict to transfer information between analyses.
        self.primary_analysis = dict()

    def build_dataloaders(self):
        """A function to build dataloader for train, validation, and test data."""
        dataloader_config = deepcopy(self.analysis_config.get('dataloader'))

        if not dataloader_config:
            self.flow_logger.warning('DataLoader configs not provided!')
            self.train_dl = None
            self.val_dl = None
            self.test_dl = None
            return

        self.train_dl, _ = build_dataloader(dataloader_config, self.train_data,
                                            self.target, self.mappings)

        self.val_dl, _ = build_dataloader(dataloader_config, self.val_data,
                                          self.target, self.mappings)

        self.test_dl, dataloader_config = build_dataloader(
            dataloader_config, self.test_data, self.target, self.mappings)

        self.analysis_config['dataloader'] = dataloader_config

    def load_data_and_targets_from_config(self, data_config: dict):
        """A function to load data and targets from data config.

        Args:
            data_config: Data config.
        """
        self.train_data, self.val_data = load_train_val_data_from_config(
            data_config)
        self.test_data = load_test_data_from_config(data_config)
        self.full_data = load_full_data_from_config(data_config)

        self.target = data_config.get('target')
        self.mappings = read_data(data_config['label_mappings'])
        self.build_dataloaders()

    def set_data_and_targets(self, train_data: Union[AnnData, AnnCollection],
                             val_data: Union[AnnData, AnnCollection],
                             test_data: Union[AnnData, AnnCollection],
                             target: Union[str, list[str]], mappings: dict):
        """A function to set data when you don't use data directly from config,
        but rather by other sources like feature subsetting, etc.

        Args:
            train_data (Union[AnnData, AnnCollection]): Training data.
            val_data (Union[AnnData, AnnCollection]): Validation data.
            target (Union[str, list[str]]): Target columns name(s).
            mappings (dict): Mapping of a column value to ids
                            eg. mappings[column_name][label2id] = {A: 1, B:2, ...}.
        """
        self.train_data = train_data
        self.val_data = val_data
        self.test_data = test_data
        self.full_data = None
        self.target = target
        self.mappings = mappings
        self.build_dataloaders()

    def evaluation_and_classification_report(self):
        """A function to evaluate the trained model and generate classification report on test data."""
        self.flow_logger.info(
            'Calculating accuracy and generating classification report on test set'
        )

        test_labels, pred_labels, pred_probabilities = self.model.get_predictions(
            self.test_dl, self.device)

        self.primary_analysis['test_labels'] = test_labels
        self.primary_analysis['pred_labels'] = pred_labels
        self.primary_analysis['pred_probabilities'] = pred_probabilities
        self.primary_analysis['mapping'] = self.mappings[
            self.target]['id2label']

        accuracy = get_accuracy(test_labels, pred_labels)

        # Generate classification report.
        generate_and_save_classification_report(
            test_labels,
            pred_labels,
            self.dirpath,
            mapping=self.mappings[self.target]['id2label'])

    def gene_analysis(self):
        """A function to perform analysis on trained model to get top genes and biomarkers."""

        self.flow_logger.info('Performing gene analysis')
        gene_analysis_path = path.join(self.dirpath, 'gene_analysis')
        gene_analyser = FeatureExtractionPipeline(
            self.analysis_config.get('gene_analysis'), gene_analysis_path,
            self.device)
        gene_analyser.set_model([self.model])
        gene_analyser.set_data_and_targets(self.train_data, self.val_data,
                                           self.target, self.mappings)

        score_matrix = gene_analyser.feature_scoring()
        self.primary_analysis['score_matrix'] = score_matrix

        top_features = gene_analyser.top_feature_extraction()
        self.primary_analysis['top_features'] = top_features

    def _perform_downstream_analysis(self, samples: str):
        """A function to perform all downstream analysis tasks on model and data.
        
        Args:
            samples ['test' | 'full']: indicates the samples to perform downstream 
                                       analysis.
        """
        downstream_analysis = self.analysis_config.get(
            f'{samples}_samples_downstream_analysis', list())
        if downstream_analysis:
            self.flow_logger.info(
                f'Performing Downstream Analysis on {samples} samples')
            samples_analysis_path = path.join(self.dirpath,
                                              f'{samples}_samples')
            os.makedirs(samples_analysis_path, exist_ok=True)

        for i, (analysis_config) in enumerate(downstream_analysis):
            try:
                self.flow_logger.info(f'Performing {analysis_config["name"]}')

                analysis_config = deepcopy(analysis_config)
                analyser, analysis_config = build_analyser(analysis_config)
                downstream_analysis[i] = analysis_config

                model = self.model if samples == 'test' else None
                dl = self.test_dl if samples == 'test' else None
                data = self.test_data if samples == 'test' else self.full_data

                analysis = analyser.generate_analysis(
                    model=model,
                    test_data=data,
                    test_dl=dl,
                    dirpath=samples_analysis_path,
                    **self.primary_analysis)

                # To be able to use any above analyses in other downstream analysis.
                if analysis:
                    self.primary_analysis[analysis_config['name']] = analysis
            except Exception as e:
                self.flow_logger.exception(e)

        if downstream_analysis:
            self.analysis_config[
                f'{samples}_samples_downstream_analysis'] = downstream_analysis

    def full_samples_downstream_anlaysis(self):
        """A function to perform downstream analysis tasks on all samples data.
        
        Note: The Model & DataLoader will not be passsed since it is assumed that
        a model is trained on the train data, so analysis by model should not
        be on full samples data.
        """
        self._perform_downstream_analysis('full')

    def test_samples_downstream_anlaysis(self):
        """A function to perform downstream analysis tasks on model and test 
        samples data.
        """
        self._perform_downstream_analysis('test')

    def get_updated_config(self) -> dict:
        """A function to return updated configs."""
        return self.analysis_config



"""This file is the implementation of simple metadataloader."""

from typing import Union

from anndata import AnnData
from anndata.experimental import AnnCollection
import numpy as np
from sklearn.preprocessing import OneHotEncoder
import torch

from scalr.nn.dataloader import SimpleDataLoader


class SimpleMetaDataLoader(SimpleDataLoader):
    """Class for simple metadataloader.
    
    Simple MetaDataLoader converts all adata values to inputs, concat specified metadata columns as onehotencoded vector
    to feature data and map target columns in metadata to output labels.

    Returns:
        PyTorch DataLoader object with (X: Tensor [batch_size, features], y: Tensor [batch_size, ]).
    """

    def __init__(self,
                 batch_size: int,
                 target: str,
                 mappings: dict,
                 metadata_col: list[str],
                 padding: int = None):
        """
        Args:
            batch_size (int): Number of samples to be loaded in each batch.
            target (str): Corresponding metadata name to be treated as training
                          objective in classification. Must be present as a column_name in `adata.obs`.
            mappings (dict): Mapping the target name to respective ids.
            metadata_col (list): List of metadata columns to be onehotencoded.
            padding (int): Padding size incase of #features < model input size.
        """
        super().__init__(batch_size=batch_size,
                         target=target,
                         mappings=mappings,
                         padding=padding)
        self.mappings = mappings
        self.metadata_col = metadata_col

        # Generating OneHotEncoder object for specified metadata_col.
        self.metadata_onehotencoder = {}
        for col in self.metadata_col:
            ohe = OneHotEncoder(handle_unknown='ignore')
            ohe.fit(np.array(sorted(mappings[col]['id2label'])).reshape(-1, 1))
            self.metadata_onehotencoder[col] = ohe

    def collate_fn(
        self,
        adata_batch: Union[AnnData, AnnCollection],
    ):
        """Given an input anndata of batch_size, the collate function creates inputs and outputs.

        Args:
            adata_batch (Union[AnnData, AnnCollection]): Anndata view object with batch_size samples.

        Returns:
            Tuple(x, y): Input to model, output from data.
        """

        # Getting x & y
        x, y = super().collate_fn(adata_batch)

        # One hot encoding requested metadata columns and adding to features data.
        for col in self.metadata_col:
            x = torch.cat(
                (x,
                 torch.as_tensor(self.metadata_onehotencoder[col].transform(
                     adata_batch.obs[col].values.reshape(-1, 1)).A,
                                 dtype=torch.float32)),
                dim=1)
        return x, y

    @classmethod
    def get_default_params(cls) -> dict:
        """Class method to get default params for model_config."""
        return dict(batch_size=1, target=None, mappings=dict())



"""This file contains an implementation of end-to-end pipeline execution."""

import argparse
import logging
import os
from os import path
import random
import sys
from time import time

from memory_profiler import memory_usage
from memory_profiler import profile
import numpy as np
import torch

from scalr.data_ingestion_pipeline import DataIngestionPipeline
from scalr.eval_and_analysis_pipeline import EvalAndAnalysisPipeline
from scalr.feature_extraction_pipeline import FeatureExtractionPipeline
from scalr.model_training_pipeline import ModelTrainingPipeline
from scalr.utils import EventLogger
from scalr.utils import FlowLogger
from scalr.utils import read_data
from scalr.utils import set_seed
from scalr.utils import write_data


def get_args():
    """A function to get the command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument('-c',
                        '--config',
                        type=str,
                        help='config.yaml file path',
                        required=True)
    parser.add_argument('-l',
                        '--log',
                        action='store_true',
                        help='flag to store logs for the experiment')
    parser.add_argument('--level',
                        type=str,
                        default='INFO',
                        help='set the level of logging')
    parser.add_argument('--logpath',
                        type=str,
                        default=False,
                        help='path to store the logs')
    parser.add_argument('-m',
                        '--memoryprofiler',
                        action='store_true',
                        help='flag to get memory usage analysis')

    args = parser.parse_args()
    return args


# Uncomment `@profile` to get line-by-line memory analysis
# @profile
def pipeline(config, dirpath, device, flow_logger, event_logger):
    """A function that configures all components of the pipeline for end-to-end execution.
    
    Args:
        config: User config.
        dirpath: Path of root directory.
        flow_logger: Object for flow logger.
        event_logger: Object for event logger.
    """
    if config.get('data'):
        # Data ingestion.
        flow_logger.info('Data Ingestion pipeline running')
        event_logger.heading1('Data Ingestion')

        data_dirpath = path.join(dirpath, 'data')
        os.makedirs(data_dirpath, exist_ok=True)

        ingest_data = DataIngestionPipeline(config['data'], data_dirpath)
        ingest_data.generate_train_val_test_split()
        ingest_data.preprocess_data()
        if not config['data'].get('label_mappings'):
            ingest_data.generate_mappings()

        config['data'] = ingest_data.get_updated_config()
        write_data(config, path.join(dirpath, 'config.yaml'))
        del ingest_data

    if config.get('feature_selection'):
        # Feature selection.
        flow_logger.info('Feature Extraction pipeline running')
        event_logger.heading1('Feature Selection')

        feature_extraction_dirpath = path.join(dirpath, 'feature_extraction')
        os.makedirs(feature_extraction_dirpath, exist_ok=True)

        extract_features = FeatureExtractionPipeline(
            config['feature_selection'], feature_extraction_dirpath, device)
        extract_features.load_data_and_targets_from_config(config['data'])

        if not config['feature_selection'].get('score_matrix'):
            extract_features.feature_subsetted_model_training()
            extract_features.feature_scoring()
        else:
            extract_features.set_score_matrix(
                read_data(config['feature_selection'].get('score_matrix')))

        extract_features.top_feature_extraction()
        config['data'] = extract_features.write_top_features_subset_data(
            config['data'])

        config['feature_selection'] = extract_features.get_updated_config()
        write_data(config, path.join(dirpath, 'config.yaml'))
        del extract_features

    if config.get('final_training'):
        # Final model training.
        flow_logger.info('Final Model Training pipeline running')
        event_logger.heading1('Final Model Training')

        model_training_dirpath = path.join(dirpath, 'model')
        os.makedirs(model_training_dirpath, exist_ok=True)

        model_trainer = ModelTrainingPipeline(
            config['final_training']['model'],
            config['final_training']['model_train_config'],
            model_training_dirpath, device)

        model_trainer.load_data_and_targets_from_config(config['data'])
        model_trainer.build_model_training_artifacts()
        model_trainer.train()

        model_config, model_train_config = model_trainer.get_updated_config()
        config['final_training']['model'] = model_config
        config['final_training']['model_train_config'] = model_train_config
        write_data(config, path.join(dirpath, 'config.yaml'))
        del model_trainer

    if config.get('analysis'):
        # Analysis of trained model.
        flow_logger.info('Analysis pipeline running')
        event_logger.heading1('Analysis')

        analysis_dirpath = path.join(dirpath, 'analysis')
        os.makedirs(analysis_dirpath, exist_ok=True)

        if config.get('final_training'):
            config['analysis']['model_checkpoint'] = path.join(
                model_training_dirpath, 'best_model')

        analyser = EvalAndAnalysisPipeline(config['analysis'], analysis_dirpath,
                                           device)
        analyser.load_data_and_targets_from_config(config['data'])

        if config['analysis'].get('model_checkpoint'):
            analyser.evaluation_and_classification_report()

            if config['analysis'].get('gene_analysis'):
                analyser.gene_analysis()

        analyser.full_samples_downstream_anlaysis()
        analyser.test_samples_downstream_anlaysis()

        config['analysis'] = analyser.get_updated_config()
        write_data(config, path.join(dirpath, 'config.yaml'))
        del analyser

    return config


if __name__ == '__main__':
    set_seed(42)
    args = get_args()

    start_time = time()

    # Parsing config.
    config = read_data(args.config)

    # Setting experiment information form config.
    dirpath = config['experiment']['dirpath']
    exp_name = config['experiment']['exp_name']
    exp_run = config['experiment']['exp_run']
    dirpath = os.path.join(dirpath, f'{exp_name}_{exp_run}')
    device = config['device']

    # Logging.
    log = args.log
    if log:
        level = getattr(logging, args.level)
        logpath = args.logpath if args.logpath else path.join(
            dirpath, 'logs.txt')
    else:
        level = logging.CRITICAL
        logpath = None

    flow_logger = FlowLogger('ROOT', level)
    flow_logger.info(f'Experiment directory: `{dirpath}`')
    if os.path.exists(dirpath):
        flow_logger.warning('Experiment directory already exists!')

    os.makedirs(dirpath, exist_ok=True)

    event_logger = EventLogger('ROOT', level, logpath)

    kwargs = dict(config=config,
                  dirpath=dirpath,
                  device=device,
                  flow_logger=flow_logger,
                  event_logger=event_logger)

    if args.memoryprofiler:
        max_memory = memory_usage((pipeline, [], kwargs),
                                  max_usage=True,
                                  interval=0.5,
                                  max_iterations=1)
    else:
        pipeline(**kwargs)

    end_time = time()
    flow_logger.info(f'Total time taken: {end_time - start_time} s')

    event_logger.heading1('Runtime Analyis')
    event_logger.info(f'Total time taken: {end_time - start_time} s')

    if args.memoryprofiler:
        flow_logger.info(f'Maximum memory usage: {max_memory} MB')
        event_logger.info(f'Maximum memory usage: {max_memory} MB')
